import numpy as np
from scipy.cluster.vq import vq

from ...enums import Metric
from .base import BaseCodec


class VQCodec(BaseCodec):
    def __init__(
        self,
        n_clusters: int,
        metric: Metric = Metric.EUCLIDEAN,
        iter: int = 100,
        n_init: int = 4,
        *args,
        **kwargs
    ):
        super(VQCodec, self).__init__(require_train=True)
        self.n_clusters = n_clusters

        # assert (
        #    metric == Metric.EUCLIDEAN
        # ), f'The distance metric `{metric.name}` is not supported yet!'
        self.metric = metric
        self._codebook = None
        self.iter = iter
        self.kmeans = None
        self.n_init = n_init

    def __hash__(self):
        return hash((self.__class__.__name__, self.n_clusters, self.metric))

    def fit(self, x: 'np.ndarray'):
        """Given training vectors, run k-means for each sub-space and create
           codewords for each sub-space.

        :param x: Training vectors with shape=(N, D) and dtype=np.float32.
        :param iter: The number of iteration for k-means
        """

        from sklearn.cluster import KMeans

        assert x.dtype == np.float32
        assert x.ndim == 2

        self.kmeans = KMeans(self.n_clusters, max_iter=self.iter, n_init=self.n_init)
        self.kmeans.fit(x)
        self._codebook = self.kmeans.cluster_centers_
        self._is_trained = True

    def partial_fit(self, x: 'np.ndarray'):
        """Given a batch of training vectors, update the internal MiniBatchKMeans.
        This method is specially designed to be used when data does not fit in memory.

        :param x: Training vectors with shape=(N, D)
        """
        assert x.ndim == 2
        if self.kmeans:
            self.kmeans.partial_fit(x)
        else:
            from sklearn.cluster import MiniBatchKMeans

            self.kmeans = MiniBatchKMeans(
                n_clusters=self.n_clusters, max_iter=self.iter
            )
            self.kmeans.partial_fit(x)

    def build_codebook(self):
        """Constructs a codebook from the current MiniBatchKmeans
        This step is not necessary if full KMeans is trained used calling `.fit`.
        """
        self._codebook = self.kmeans.cluster_centers_
        self._is_trained = True

    def encode(self, x: 'np.ndarray'):
        """Encodes each row of the input array `x` it's closest cluster id."""
        self._check_trained()
        assert x.dtype == np.float32
        assert x.ndim == 2

        codes, _ = vq(x, self.codebook)
        return codes

    def decode(self, x: 'np.ndarray'):
        return None

    @property
    def codebook(self):
        self._check_trained()
        return self._codebook



import os
import platform
import shutil
import time
from pathlib import Path
from typing import Optional, Union

from filesplit.merge import Merge
from filesplit.split import Split
from loguru import logger

ignored_extn = ['.DS_Store']


def get_size(input: Path) -> float:
    import os

    return os.stat(str(input)).st_size / (1024 * 1024)


def make_archive(input: Path, output_name: str) -> Path:
    """
    This function will create a zip archive of the input file (tmp.zip) at the
    same folder of input path.
    """
    output_path = shutil.make_archive(
        os.path.join(str(input.parent), output_name),
        'zip',
        str(input.parent),
        str(input.name),
    )
    return Path(output_path)


class Uploader:
    def __init__(self, size_limit=1024, client=None):
        """
        This class create a filesplit object to split the file into small pieces and
        upload them on to hubble.
        :params size_limit: The max size of split files.
        :params client: hubble client used for uploading.
        """
        self.size_limit = size_limit
        self.client = client

    def upload_file(
        self, input: Path, target_name: str, type: str, cell_id: Union[int, str]
    ):
        logger.info(f'Start to upload single file: {input} to hubble ...')
        size = get_size(input)
        if size > self.size_limit:
            split_list = self._split_file(input)
            self.upload_directory(split_list, target_name, type, cell_id, merge=False)
            shutil.rmtree(split_list)
        else:
            if self._check_exists(target_name, type, input.name):
                return
            self._upload_hubble(input, target_name, type, input.name, cell_id)

    def upload_directory(
        self,
        input: Path,
        target_name: str,
        type: str,
        cell_id: Union[int, str],
        merge: bool = True,
    ):
        def _upload():
            if self._check_exists(target_name, type, str(idx) + '.zip'):
                return

            Path.mkdir(input.parent / str(idx))
            for f in split_list:
                shutil.copy(f, input.parent / str(idx))
            output_path = make_archive(input.parent / str(idx), str(idx) + '.zip')
            self._upload_hubble(
                output_path, target_name, type, str(idx) + '.zip', cell_id
            )
            Path(output_path).unlink()
            shutil.rmtree(input.parent / str(idx))

        logger.info(f'Start to upload directory: {input} to hubble ...')
        if merge:
            size_list = list(
                zip(list(input.iterdir()), [get_size(f) for f in list(input.iterdir())])
            )
            sorted_size_list = sorted(size_list, key=lambda x: x[1])

            split_list = []
            total_size = 0
            idx = 0

            for file_name, file_size in sorted_size_list:
                for extn in ignored_extn:
                    if extn in str(file_name):
                        continue
                if total_size + file_size > self.size_limit:
                    if len(split_list) == 0:
                        raise Exception(
                            f'The smallest file: {file_size} is bigger '
                            f'than size_limit. Please set a larger value '
                            f'of size_limit, now is {self.size_limit}MB.'
                        )
                    _upload()
                    idx += 1
                    total_size = 0
                    split_list = [file_name]
                else:
                    split_list.append(file_name)
                    total_size += file_size
            if len(split_list) > 0:
                _upload()
        else:
            for idx, file_name in enumerate(list(input.glob('*'))):
                if self._check_exists(target_name, type, str(file_name.name)):
                    continue
                self._upload_hubble(
                    file_name, target_name, type, str(file_name.name), cell_id
                )

    def archive_and_upload(
        self,
        target_name: str,
        type: str,
        file_name: str,
        cell_id: Union[int, str],
        root_path: Path,
        upload_folder: str,
    ):
        if self._check_exists(target_name, type, file_name):
            return
        upload_file = shutil.make_archive(
            os.path.join(str(root_path), f'{target_name}_{type}'),
            'zip',
            str(root_path),
            upload_folder,
        )

        logger.info(
            f'Start to upload: {upload_file} to hubble. '
            f'[target_name: {target_name}, '
            f'type: {type}, '
            f'file_name: {file_name}, '
            f'cell_id: {cell_id}].'
        )

        self.client.upload_artifact(
            f=upload_file,
            metadata={
                'name': target_name,
                'type': type,
                'file_name': file_name,
                'cell': cell_id,
            },
        )
        Path(upload_file).unlink()

    def _check_exists(self, target_name: str, type: str, file_name: str) -> bool:
        art_list = self.client.list_artifacts(
            filter={
                'metaData.name': target_name,
                'metaData.type': f'{type}',
                'metaData.file_name': f'{file_name}',
            }
        )
        if len(art_list['data']) != 0:
            logger.info(
                f'[target_name: {target_name}, type: {type}, file_name: {file_name}] '
                f'already exists on hubble, will skip it ...'
            )
            return True
        else:
            return False

    def _split_file(self, input: Path) -> Path:
        output_dir = input / f'{input}_split'
        if output_dir.exists():
            logger.info(
                f'Origin file: {str(input)} has already been split to: {output_dir}, will skip ...'
            )
            return output_dir
        Path.mkdir(output_dir)
        Split(str(input), str(output_dir)).bysize(size=self.size_limit * 1024 * 1024)
        num_files = len(list(output_dir.glob('*')))
        logger.info(
            f'Origin file: {str(input)} has been split '
            f'into {num_files} parts. Output file: {output_dir}'
        )
        return output_dir

    def _upload_hubble(
        self,
        upload_file: Path,
        target_name: str,
        type: str,
        file_name: str,
        cell_id: Union[str, int],
    ):

        logger.info(
            f'Start to upload: {upload_file} to hubble. '
            f'[target_name: {target_name}, '
            f'type: {type}, '
            f'file_name: {file_name}, '
            f'cell_id: {cell_id}].'
        )

        start_time = time.time()
        failed_times = 0

        while True:
            try:
                self.client.upload_artifact(
                    f=str(upload_file),
                    metadata={
                        'name': target_name,
                        'type': type,
                        'file_name': file_name,
                        'cell': cell_id,
                    },
                    show_progress=True,
                )
                break
            except Exception as e:
                logger.info(e)
                failed_times += 1
                if failed_times == 3:
                    logger.info(
                        f'Tried more than 3 times to upload {upload_file}, type is: {type}, will exist...'
                    )
                    return
                else:
                    continue

        logger.info(
            f'Takes {time.time() - start_time} seconds to upload {upload_file}.'
        )


class Merger:
    def __init__(self, restore_path, client):
        """
        This class creates an object to download and merge the split files from hubble.
        :param restore_path: tmp directory for downloading and merging files.
        :param client: hubble client used for merging files.
        """
        self.restore_path = restore_path
        self.restore_path.mkdir(parents=True)
        self.client = client

    def merge_file(self, inputdir: Path, outputdir: Path, outputfilename: Path):
        Merge(
            inputdir=str(inputdir),
            outputdir=str(outputdir),
            outputfilename=str(outputfilename),
        ).merge()

    def get_artifact_ids(self, art_list, type: str, cell_id: Optional[int] = None):
        ids = [
            [
                art['_id'],
                art['metaData']['type'],
                art['metaData']['file_name'],
                art['metaData']['cell'],
            ]
            for art in art_list['data']
            if type == art['metaData']['type']
        ]
        if cell_id:
            ids = [item for item in ids if int(item[3]) == cell_id]
        ids = [[item[0], item[1], item[2]] for item in ids]
        return ids

    def download(self, ids, download_folder):
        Path.mkdir(self.restore_path / download_folder)

        for ids, type, file_name in ids:
            self.client.download_artifact(
                id=ids,
                f=str(self.restore_path / download_folder / file_name),
                show_progress=True,
            )



import abc
from typing import List, Optional, Union

import numpy as np

from ...enums import ExpandMode, Metric
from ...helper import str2dtype


class BaseIndex(abc.ABC):
    def __init__(
        self,
        dim: int,
        dtype: Union[np.dtype, str] = np.float32,
        metric: Metric = Metric.COSINE,
        initial_size: Optional[int] = None,
        expand_step_size: int = 10240,
        expand_mode: ExpandMode = ExpandMode.STEP,
        *args,
        **kwargs
    ):
        assert expand_step_size > 0
        self.initial_size = initial_size or expand_step_size

        self.expand_step_size = expand_step_size
        self.expand_mode = expand_mode

        self.dim = dim
        self.dtype = str2dtype(dtype) if isinstance(dtype, str) else dtype
        self.metric = metric

        self._size = 0
        self._capacity = self.initial_size

    @property
    def capacity(self) -> int:
        return self._capacity

    @property
    def size(self):
        return self._size

    @abc.abstractmethod
    def add_with_ids(self, x: np.ndarray, ids: List[int], **kwargs):
        ...

    @abc.abstractmethod
    def delete(self, ids: List[int]):
        ...

    @abc.abstractmethod
    def update_with_ids(self, x: np.ndarray, ids: List[int], **kwargs):
        ...

    def reset(self, capacity: Optional[int] = None):
        self._size = 0
        self._capacity = capacity or self.initial_size



from typing import Optional

import numpy as np

from .base import BaseCodec


class ProjectorCodec(BaseCodec):
    """Implementation of Projector.

    :param n_components: number of components to keep.
    :param whiten: when True (False by default) the components_ vectors are multiplied
                    by the square root of n_samples and then divided by the singular
                    values to ensure uncorrelated outputs with unit component-wise variances.
    :param svd_solver:
            If auto: The solver is selected by a default policy based on X.shape and
            n_components: if the input data is larger than 500x500 and the number of
            components to extract is lower than 80% of the smallest dimension of the
            data, then the more efficient ‘randomized’ method is enabled. Otherwise
            the exact full SVD is computed and optionally truncated afterwards.

            If full: run exact full SVD calling the standard LAPACK solver via scipy.
            linalg.svd and select the components by postprocessing.

            If arpack: run SVD truncated to n_components calling ARPACK solver via
            scipy.sparse.linalg.svds. It requires strictly 0 < n_components < min(X.shape).
    """

    def __init__(
        self,
        dim: int,
        n_components: int = 128,
        whiten: Optional[bool] = False,
        svd_solver: Optional[str] = 'auto',
    ):
        super(ProjectorCodec, self).__init__(require_train=True)
        self.dim = dim
        self.n_components = n_components
        assert self.dim >= self.n_components, (
            f'the dimension after projector should be less than original dimension, got '
            f'original dimension: {self.dim} and projector dimension: {self.n_components}'
        )

        self.whiten = whiten
        self.svd_solver = svd_solver

        self.pca = None

    def __hash__(self):
        return hash(
            (
                self.__class__.__name__,
                self.dim,
                self.n_components,
                self.whiten,
                self.svd_solver,
            )
        )

    def fit(self, x: 'np.ndarray'):
        """Train projector model

        :param x: Training vectors with shape=(N, D)
        """
        assert x.ndim == 2
        assert (
            x.shape[1] == self.dim,
        ), 'dimension of input data must be equal to "dim"'
        assert (
            x.shape[0] > self.n_components
        ), 'number of input data must be larger than or equal to n_components'

        if self.pca is None:
            from sklearn.decomposition import PCA

            self.pca = PCA(
                n_components=self.n_components,
                whiten=self.whiten,
                svd_solver=self.svd_solver,
            )

        self.pca.fit(x)
        self._is_trained = True

    def partial_fit(self, x: 'np.ndarray'):
        """Given a batch of training vectors, update the internal projector.
        This method is specially designed to be used when data does not fit in memory.

        :param x: Training vectors with shape=(N, D)
        """

        assert x.ndim == 2
        assert x.shape[1] == self.dim, 'dimension of input data must be equal to "dim"'
        assert (
            x.shape[0] > self.n_components
        ), 'number of input data must be larger than or equal to n_components'

        if self.pca is None:
            from sklearn.decomposition import IncrementalPCA

            self.pca = IncrementalPCA(
                n_components=self.n_components,
                whiten=self.whiten,
            )

        self.pca.partial_fit(x)
        self._is_trained = True

    def encode(self, x: 'np.ndarray'):
        """Encode input vectors using projector.

        :param x: Input vectors with shape=(N, D)
        :return: np.ndarray: transformed vectors using projector.
        """
        assert x.ndim == 2
        assert x.shape[1] == self.dim, 'dimension of input data must be equal to "dim"'

        return self.pca.transform(x)

    def decode(self, x: 'np.ndarray'):
        """Given transformed vectors, reconstruct original D-dimensional vectors
        approximately.

        :param x: vectors with shape=(N, self.n_components).
        :return: Reconstructed vectors with shape=(N, D)
        """
        assert x.ndim == 2
        assert x.shape[1] == self.n_components

        return self.pca.inverse_transform(x)

    @property
    def components(self):
        """Principal axes in feature space, representing the directions of maximum
        variance in the data.
        """
        self._check_trained()
        return self.pca.components_

    @property
    def explained_variance_ratio(self):
        """Percentage of variance explained by each of the selected components."""
        self._check_trained()
        return self.pca.explained_variance_ratio_

    @property
    def mean(self):
        """Per-feature empirical mean."""
        self._check_trained()
        return self.pca.mean_

    @property
    def var(self):
        """Per-feature empirical variance"""
        self._check_trained()
        return self.pca.var_



import pickle
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path


class BaseCodec(ABC):
    def __init__(self, require_train: bool = True):
        self.require_train = require_train
        self._is_trained = False if require_train else True

    @abstractmethod
    def fit(self, *args, **kwargs):
        pass

    @abstractmethod
    def encode(self):
        pass

    @abstractmethod
    def decode(self):
        pass

    def dump(self, target_path: 'Path'):
        pickle.dump(self, target_path.open('wb'), protocol=4)

    @staticmethod
    def load(from_path: 'Path'):
        return pickle.load(from_path.open('rb'))

    @property
    def is_trained(self):
        return self._is_trained

    def _check_trained(self):
        assert self.is_trained is True, f'{self.__class__.__name__} requires training'



from enum import IntEnum


class BetterEnum(IntEnum):
    """The base class of Enum."""

    def __str__(self):
        return self.name

    @classmethod
    def from_string(cls, s: str):
        """
        Parse the enum from a string.
        :param s: string representation of the enum value
        :return: enum value
        """
        try:
            return cls[s.upper()]
        except KeyError:
            raise ValueError(
                f'{s.upper()} is not a valid enum for {cls!r}, must be one of {list(cls)}'
            )


class Metric(BetterEnum):
    EUCLIDEAN = 1
    INNER_PRODUCT = 2
    COSINE = 3


class ExpandMode(BetterEnum):
    STEP = 1
    DOUBLE = 2
    ADAPTIVE = 3



from typing import Dict

LOGICAL_OPERATORS = {'$and': 'AND', '$or': 'OR'}

COMPARISON_OPERATORS = {
    '$lt': '<',
    '$gt': '>',
    '$lte': '<=',
    '$gte': '>=',
    '$eq': '=',
    '$neq': '!=',
}

MEMBERSHIP_OPERATORS = {'$in': 'IN', '$nin': 'NOT IN'}


def _sql_parsing(data, default_logic: str = 'AND'):
    """
    :param data: JSON Object (dict).
    :param parameters: dict.
    :return: where clause (str) built from data
    """
    where_clause = ''
    parameters = []

    if isinstance(data, dict):
        for i, (key, value) in enumerate(data.items()):
            if key in LOGICAL_OPERATORS:
                clause, params = _sql_parsing(
                    value, default_logic=LOGICAL_OPERATORS[key]
                )
                if i == 0:
                    where_clause += clause
                else:
                    where_clause += f' {LOGICAL_OPERATORS[key]} {clause}'
                parameters.extend(params)
            elif key.startswith('$'):
                raise ValueError(
                    f'The operator {key} is not supported yet, please double check the given filters!'
                )
            else:
                if i > 0:
                    where_clause += f' {default_logic} '

                items = list(value.items())

                if len(items) == 0:
                    raise ValueError(f'The query express is illegal: {data}')
                elif len(items) > 1:
                    clause_list, params_list = [], []

                    for op, val in items:
                        _clause, _params = _sql_parsing({key: {op: val}})
                        clause_list.append(_clause)
                        params_list.extend(_params)

                    where_clause += f' AND '.join(clause_list)
                    parameters.extend(params_list)
                else:
                    op, val = items[0]
                    if op in LOGICAL_OPERATORS:
                        clause, params = _sql_parsing(
                            val, default_logic=LOGICAL_OPERATORS[op]
                        )
                        where_clause += clause
                        parameters.extend(params)
                    elif op in COMPARISON_OPERATORS:
                        parameters.append(val)
                        where_clause += f'({key} {COMPARISON_OPERATORS[op]} ?)'
                    elif op in MEMBERSHIP_OPERATORS:
                        parameters.extend(val)
                        where_clause += f'({key} {MEMBERSHIP_OPERATORS[op]}({", ".join(["?"]*len(val))}))'
                    else:
                        raise ValueError(
                            f'The operator {op} is not supported yet, please double check the given filters!'
                        )
    elif isinstance(data, list):
        clause_list, params_list = [], []
        for d in data:
            _clause, _params = _sql_parsing(d)
            clause_list.append(_clause)
            params_list.extend(_params)
        where_clause += '(' + f' {default_logic} '.join(clause_list) + ')'
        parameters.extend(params_list)

    elif isinstance(data, str):
        return data, parameters
    else:
        raise ValueError(f'The query express is illegal: {data}')
    return where_clause, tuple(parameters)


class Filter(object):
    """A class to parse query language to SQL where clause."""

    def __init__(self, tree_data: Dict = {}):
        self.tree_data = tree_data

    def parse_where_clause(self):
        return _sql_parsing(self.tree_data or {})



import math
import os.path
from functools import wraps
from pathlib import Path
from typing import TYPE_CHECKING, List, Optional, Union

import numpy as np
from loguru import logger

from annlite.hnsw_bind import Index

from ....enums import Metric
from ....math import l2_normalize
from ..base import BaseIndex

if TYPE_CHECKING:
    from ...codec.base import BaseCodec


def pre_process(f):
    @wraps(f)
    def pre_processed(self: 'HnswIndex', x: np.ndarray, *args, **kwargs):
        if x.ndim == 1:
            x = x.reshape((1, -1))
        if x.dtype != self.dtype:
            x = x.astype(self.dtype)

        if self.normalization_enable:
            x = l2_normalize(x)

        if self.pq_enable:
            if not self.pq_codec.is_trained:
                raise RuntimeError(
                    'Please train the PQ before using HNSW quantization backend'
                )
            elif not self._set_backend_pq:
                self._index.loadPQ(self.pq_codec)
                self._set_backend_pq = True
            kwargs['pre_process_dtables'] = self.pq_codec.get_dist_mat(x)
            x = self.pq_codec.encode(x)

            assert kwargs['pre_process_dtables'].dtype == 'float32'
            assert kwargs['pre_process_dtables'].flags['C_CONTIGUOUS']
            return f(self, x, *args, **kwargs)
        else:
            return f(self, x, *args, **kwargs)

    return pre_processed


class HnswIndex(BaseIndex):
    def __init__(
        self,
        dim: int,
        dtype: np.dtype = np.float32,
        metric: Metric = Metric.COSINE,
        ef_construction: int = 200,
        ef_search: int = 50,
        max_connection: int = 16,
        pq_codec: Optional['BaseCodec'] = None,
        index_file: Optional[Union[str, Path]] = None,
        **kwargs,
    ):
        """
        :param dim: The dimensionality of vectors to index
        :param index_file: A file-like object or a string containing a file name.
        :param metric: Distance metric type, can be 'euclidean', 'inner_product', or 'cosine'
        :param ef_construction: the size of the dynamic list for the nearest neighbors (used during the building).
        :param ef_search: the size of the dynamic list for the nearest neighbors (used during the search).
        :param max_connection: The number of bi-directional links created for every new element during construction.
                    Reasonable range for M is 2-100.
        """
        super().__init__(dim, dtype=dtype, metric=metric, **kwargs)

        self.ef_construction = ef_construction
        self.ef_search = ef_search
        self.max_connection = max_connection
        self.pq_codec = pq_codec
        self._set_backend_pq = False
        self.index_file = index_file

        self._init_hnsw_index()

    def _init_hnsw_index(self):
        self._index = Index(space=self.space_name, dim=self.dim)
        if self.index_file:
            if os.path.exists(self.index_file):
                logger.info(
                    f'indexer will be loaded from {self.index_file}',
                )
                self.load(self.index_file)
            else:
                raise FileNotFoundError(
                    f'index path: {self.index_file} does not exist',
                )
        else:
            if self.pq_codec is not None and self.pq_codec.is_trained:
                self._index.init_index(
                    max_elements=self.capacity,
                    ef_construction=self.ef_construction,
                    M=self.max_connection,
                    pq_codec=self.pq_codec,
                )
                self._set_backend_pq = True
            else:
                self._index.init_index(
                    max_elements=self.capacity,
                    ef_construction=self.ef_construction,
                    M=self.max_connection,
                    pq_codec=None,
                )
                self._set_backend_pq = False

        self._index.set_ef(self.ef_search)

    def load(self, index_file: Union[str, Path]):
        self._index.load_index(str(index_file))
        if self.pq_codec:
            self._index.loadPQ(self.pq_codec)

    def dump(self, index_file: Union[str, Path]):
        self._index.save_index(str(index_file))

    @pre_process
    def add_with_ids(
        self,
        x: 'np.ndarray',
        ids: List[int],
        # kwargs maybe used by pre_process
        pre_process_dtables=None,
    ):
        max_id = max(ids) + 1
        if max_id > self.capacity:
            expand_steps = math.ceil(max_id / self.expand_step_size)
            self._expand_capacity(expand_steps * self.expand_step_size)

        self._index.add_items(x, ids=ids, dtables=pre_process_dtables)

    @pre_process
    def search(
        self,
        query: 'np.ndarray',
        limit: int = 10,
        indices: Optional['np.ndarray'] = None,
        # kwargs maybe used by pre_process
        pre_process_dtables=None,
    ):
        ef_search = max(self.ef_search, limit)
        self._index.set_ef(ef_search)

        if indices is not None:
            # TODO: add a smart strategy to speed up this case (bruteforce search would be better)
            if len(indices) < limit:
                limit = len(indices)
            ids, dists = self._index.knn_query_with_filter(
                query, filters=indices, k=limit, dtables=pre_process_dtables
            )
        else:
            ids, dists = self._index.knn_query(
                query, k=limit, dtables=pre_process_dtables
            )

        # convert squared l2 into euclidean distance
        if self.metric == Metric.EUCLIDEAN:
            dists = np.sqrt(dists)

        return dists[0], ids[0]

    def delete(self, ids: List[int]):
        for i in ids:
            self._index.mark_deleted(i)

    def update_with_ids(self, x: 'np.ndarray', ids: List[int], **kwargs):
        raise RuntimeError(
            f'the update operation is not allowed for {self.__class__.__name__}!'
        )

    def _expand_capacity(self, new_capacity: int):
        self._capacity = new_capacity
        self._index.resize_index(new_capacity)
        logger.debug(
            f'HNSW index capacity is expanded by {self.expand_step_size}',
        )

    def reset(self, capacity: Optional[int] = None):
        super().reset(capacity=capacity)
        self._init_hnsw_index()

    @property
    def size(self):
        return self._index.element_count

    @property
    def space_name(self):
        if self.metric == Metric.EUCLIDEAN:
            return 'l2'
        elif self.metric == Metric.INNER_PRODUCT:
            return 'ip'
        return 'cosine'

    @property
    def pq_enable(self):
        return self.pq_codec is not None

    @property
    def normalization_enable(self):
        return self.metric == Metric.COSINE



# distutils: language = c++

import numpy as np

cimport cython
from libc.stdint cimport (
    int8_t,
    int16_t,
    int32_t,
    int64_t,
    uint8_t,
    uint16_t,
    uint32_t,
    uint64_t,
)
from libcpp.vector cimport vector

ctypedef fused any_int:
    uint8_t
    uint16_t
    uint32_t
    uint64_t
    int8_t
    int16_t
    int32_t
    int64_t

@cython.boundscheck(False)
@cython.wraparound(False)
cdef inline dist_pqcode_to_codebook(long M,const float[:,:] adtable, any_int[:] pq_code):
    """Compute the distance between each codevector and the pq_code of a query.

    :param M: Number of sub-vectors in the original feature space.
    :param adtable: 2D Memoryview[float] containing precomputed Asymmetric Distances.
    :param pq_code: 1D Memoriview[any_int] containing a pq code.

    :return: Distance between pq code and query according to the Asymmetric Distance table.

    """
    cdef:
        float dist = 0
        int m

    for m in range(M):
        dist += adtable[m, pq_code[m]]

    return dist


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef dist_pqcodes_to_codebooks(const float[:,:] adtable, any_int[:,:] pq_codes):
    """
    Compute the distance between each row in pq_codes and each codevector using a adtable.

    :param adtable: 2D Memoryview of precomputed Asymmetric Distances.
    :param pq_codes: 2D Memoryview of pq_codes.

    :return: List of Asymmetric Distances distances between pq_codes and the query.

    This function is equivalent to:
    '''
        dists = np.zeros((N, )).astype(np.float32)
        for n in range(N):
            for m in range(M):
                dists[n] += self.adtable[m][codes[n][m]]
    '''

    """

    cdef:
        int m
        int N = pq_codes.shape[0]
        int M = pq_codes.shape[1]
        vector[float] dists

    for n in range(N):
        dists.push_back(dist_pqcode_to_codebook(M, adtable, pq_codes[n,:]))

    return dists


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef precompute_adc_table(const float[:] query,
                                      long d_subvector,
                                      long n_clusters,
                                      const float[:,:,:] codebooks):
    """
    Compute the  Asymmetric Distance Table between a query and a PQ space.

    :param query: Memoryview a query in the original feature space (not a pqcode).
    :param d_subvector: Number of dimensions in a subvector.
    :param n_clusters: Number of clusters per sub-space (number of prototypes per sub-space).
    :param codebooks: Memoryview containing the learned codevectors for each slice.
        This is a 3D view with (slice index, prototype index, vector values).

    :return: Memoryview with a 2D matrix containing the Asymmetric Distance Computation.

    This function is equivalent to
        '''
        def numpy_adc_table(query, n_subvectors, n_clusters, d_subvector, codebooks):
        adtable = np.empty((n_subvectors, n_clusters), dtype=np.float32)
        for m in range(n_subvectors):
            query_sub = query[m * d_subvector: (m + 1) * d_subvector]
            adtable[m, :] = np.linalg.norm(codebooks[m] - query_sub, axis=1) ** 2

        return adtable
        '''
    But avoids generating views and calling numpy functions.
    """

    cdef:
        int D = len(query)
        int M = int(D/d_subvector)
        int n_subvectors = int(D/d_subvector)
        int m, i, k, ind_prototype, j
        float[:, ::1] adtable = np.empty((M, n_clusters), dtype=np.float32)
        float[:] query_subvec = np.empty(d_subvector, dtype=np.float32)
        float[:] query_subcodeword = np.empty(d_subvector, dtype=np.float32)
        float dist_subprototype_to_subquery, coord_j

    for m in range(n_subvectors):

        # load m'th subquery
        i = 0
        for k in range(m * d_subvector, (m + 1) * d_subvector):
            query_subvec[i] = query[k]
            i += 1

        for ind_prototype in range(n_clusters):

            # load prototype ind_prototype for the m'th subspace
            for i in range(d_subvector):
                query_subcodeword[i] = codebooks[m, ind_prototype, i]

            # compute the distance between subprototype and subquery
            dist_subprototype_to_subquery = 0.
            for j in range(d_subvector):
                coord_j = query_subcodeword[j] - query_subvec[j]
                dist_subprototype_to_subquery += coord_j * coord_j

            adtable[m, ind_prototype] = dist_subprototype_to_subquery

    return adtable

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef batch_precompute_adc_table(const float[:, :] queries,
                                      long d_subvector,
                                      long n_clusters,
                                      const float[:,:,:] codebooks):
    """
    Compute the  Asymmetric Distance Table between a query and a PQ space.

    :param query: Memoryview a query in the original feature space (not a pqcode).
    :param d_subvector: Number of dimensions in a subvector.
    :param n_clusters: Number of clusters per sub-space (number of prototypes per sub-space).
    :param codebooks: Memoryview containing the learned codevectors for each slice.
        This is a 3D view with (slice index, prototype index, vector values).

    :return: Memoryview with a 2D matrix containing the Asymmetric Distance Computation.

    This function is equivalent to
        '''
        def numpy_adc_table(query, n_subvectors, n_clusters, d_subvector, codebooks):
        adtable = np.empty((n_subvectors, n_clusters), dtype=np.float32)
        for m in range(n_subvectors):
            query_sub = query[m * d_subvector: (m + 1) * d_subvector]
            adtable[m, :] = np.linalg.norm(codebooks[m] - query_sub, axis=1) ** 2

        return adtable
        '''
    But avoids generating views and calling numpy functions.
    """

    cdef:
        int N = queries.shape[0]
        int D = queries.shape[1]
        int n_subvectors = int(D/d_subvector)
        int m, i, k, ind_prototype, j
        float[:, :, :] adtable = np.empty((N, n_subvectors, n_clusters), dtype=np.float32)
        float[:] query_subvec = np.empty(d_subvector, dtype=np.float32)
        float[:] query_subcodeword = np.empty(d_subvector, dtype=np.float32)
        float dist_subprototype_to_subquery, coord_j

    for index in range(N):
        for m in range(n_subvectors):

            # load m'th subquery
            i = 0
            for k in range(m * d_subvector, (m + 1) * d_subvector):
                query_subvec[i] = queries[index, k]
                i += 1

            for ind_prototype in range(n_clusters):

                # load prototype ind_prototype for the m'th subspace
                for i in range(d_subvector):
                    query_subcodeword[i] = codebooks[m, ind_prototype, i]

                # compute the distance between subprototype and subquery
                dist_subprototype_to_subquery = 0.
                for j in range(d_subvector):
                    coord_j = query_subcodeword[j] - query_subvec[j]
                    dist_subprototype_to_subquery += coord_j * coord_j

                adtable[index, m, ind_prototype] = dist_subprototype_to_subquery

    return adtable

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef batch_precompute_adc_table_ip(const float[:, :] queries,
                                      long d_subvector,
                                      long n_clusters,
                                      const float[:,:,:] codebooks):
    """
    Compute the  Asymmetric Distance Table between a query and a PQ space.

    :param query: Memoryview a query in the original feature space (not a pqcode).
    :param d_subvector: Number of dimensions in a subvector.
    :param n_clusters: Number of clusters per sub-space (number of prototypes per sub-space).
    :param codebooks: Memoryview containing the learned codevectors for each slice.
        This is a 3D view with (slice index, prototype index, vector values).

    :return: Memoryview with a 2D matrix containing the Asymmetric Distance Computation.

    This function is equivalent to
        '''
        def numpy_adc_table(query, n_subvectors, n_clusters, d_subvector, codebooks):
        adtable = np.empty((n_subvectors, n_clusters), dtype=np.float32)
        for m in range(n_subvectors):
            query_sub = query[m * d_subvector: (m + 1) * d_subvector]
            adtable[m, :] = np.linalg.norm(codebooks[m] - query_sub, axis=1) ** 2

        return adtable
        '''
    But avoids generating views and calling numpy functions.
    """

    cdef:
        int N = queries.shape[0]
        int D = queries.shape[1]
        int n_subvectors = int(D/d_subvector)
        int m, i, k, ind_prototype, j
        float[:, :, :] adtable = np.empty((N, n_subvectors, n_clusters), dtype=np.float32)
        float[:] query_subvec = np.empty(d_subvector, dtype=np.float32)
        float[:] query_subcodeword = np.empty(d_subvector, dtype=np.float32)
        float dist_subprototype_to_subquery

    for index in range(N):
        for m in range(n_subvectors):

            # load m'th subquery
            i = 0
            for k in range(m * d_subvector, (m + 1) * d_subvector):
                query_subvec[i] = queries[index, k]
                i += 1

            for ind_prototype in range(n_clusters):

                # load prototype ind_prototype for the m'th subspace
                for i in range(d_subvector):
                    query_subcodeword[i] = codebooks[m, ind_prototype, i]

                # compute the distance between subprototype and subquery
                dist_subprototype_to_subquery = 0.
                for j in range(d_subvector):
                    dist_subprototype_to_subquery += (query_subcodeword[j] * query_subvec[j])

                adtable[index, m, ind_prototype] = dist_subprototype_to_subquery

    return adtable



import threading
import time
import traceback
import warnings
from threading import Thread
from typing import Dict, List, Optional, Tuple, Union

from docarray import Document, DocumentArray
from jina import Executor, requests
from jina.logging.logger import JinaLogger

INDEX_BATCH_SIZE = 1024


class AnnLiteIndexer(Executor):
    """A simple indexer that wraps the AnnLite indexer and adds a simple interface for indexing and searching.

    :param n_dim: Dimensionality of vectors to index
    :param metric: Distance metric type. Can be 'euclidean', 'inner_product', or 'cosine'
    :param limit: Number of results to get for each query document in search
    :param n_components: Number of components to use for dimensionality reduction
    :param match_args: the arguments to `DocumentArray`'s match function
    :param data_path: the workspace of the AnnLiteIndexer but not support when shards > 1.
    :param ef_construction: The construction time/accuracy trade-off
    :param ef_search: The query time accuracy/speed trade-off
    :param max_connection: The maximum number of outgoing connections in the
        graph (the "M" parameter)
    :param include_metadata: If True, return the document metadata in response
    :param index_access_paths: Default traversal paths on docs
            (used for indexing, delete and update), e.g. '@r', '@c', '@r,c'
    :param search_access_paths: Default traversal paths on docs
    (used for search), e.g. '@r', '@c', '@r,c'
    :param columns: A list or dict of column names to index.
    :param dim: Deprecated, use n_dim instead
    """

    def __init__(
        self,
        n_dim: int = 0,
        metric: str = 'cosine',
        limit: int = 10,
        n_components: Optional[int] = None,
        match_args: Optional[Dict] = None,
        data_path: Optional[str] = None,
        ef_construction: Optional[int] = None,
        ef_search: Optional[int] = None,
        max_connection: Optional[int] = None,
        include_metadata: bool = True,
        index_access_paths: str = '@r',
        search_access_paths: str = '@r',
        columns: Optional[Union[List[Tuple[str, str]], Dict[str, str]]] = None,
        dim: int = None,
        *args,
        **kwargs,
    ):

        super().__init__(*args, **kwargs)
        self.logger = JinaLogger(self.__class__.__name__)

        n_dim = n_dim or dim
        if not n_dim:
            raise ValueError('Please specify the dimension of the vectors to index!')

        self.n_components = n_components
        self.metric = metric
        self.match_args = match_args or {}
        self.include_metadata = include_metadata
        if limit:
            self.match_args.update({'limit': limit})

        self.index_access_paths = index_access_paths
        if 'index_traversal_paths' in kwargs:
            warnings.warn(
                f'`index_traversal_paths` is deprecated. Use `index_access_paths` instead.'
            )
            self.index_access_paths = kwargs['index_traversal_paths']

        self.search_access_paths = search_access_paths
        if 'search_traversal_paths' in kwargs:
            warnings.warn(
                f'`search_traversal_paths` is deprecated. Use `search_access_paths` instead.'
            )
            self.search_access_paths = kwargs['search_traversal_paths']

        self._data_buffer = DocumentArray()
        self._index_batch_size = INDEX_BATCH_SIZE
        self._max_length_queue = 2 * self._index_batch_size
        self._index_lock = threading.Lock()

        self.logger = JinaLogger(getattr(self.metas, 'name', self.__class__.__name__))

        if getattr(self.runtime_args, 'shards', 1) > 1 and data_path:
            raise ValueError(
                '`data_path` is not supported when shards > 1, please use `workspace` instead'
            )

        config = {
            'n_dim': n_dim,
            'n_components': n_components,
            'metric': metric,
            'ef_construction': ef_construction,
            'ef_search': ef_search,
            'max_connection': max_connection,
            'data_path': data_path or self.workspace or './workspace',
            'columns': columns,
        }
        self._index = DocumentArray(storage='annlite', config=config)

        # start indexing thread in background to group indexing requests
        # together and perform batch indexing at once
        self._start_index_loop()

    @requests(on='/index')
    def index(
        self, docs: Optional[DocumentArray] = None, parameters: dict = {}, **kwargs
    ):
        """Index new documents

        :param docs: the Documents to index
        :param parameters: dictionary with options for indexing
        Keys accepted:
            - 'access_paths': traversal paths on docs, e.g. '@r', '@c', '@r,c'
        """

        if not docs:
            return

        access_paths = parameters.get('access_paths', self.index_access_paths)
        flat_docs = docs[access_paths]
        if len(flat_docs) == 0:
            return

        while len(self._data_buffer) >= self._max_length_queue:
            time.sleep(0.001)

        with self._index_lock:
            self._data_buffer.extend(flat_docs)

    def _start_index_loop(self):
        """Start the indexing loop in background.

        This loop is responsible for batch indexing the documents in the buffer.
        """

        def _index_loop():
            try:
                while True:
                    # if the buffer is none, will break the loop
                    if self._data_buffer is None:
                        break

                    # if the buffer is empty, will wait for new documents to be added
                    if len(self._data_buffer) == 0:
                        time.sleep(0.1)  # sleep for 100ms
                        continue

                    # acquire the lock to prevent threading issues
                    with self._index_lock:
                        batch_docs = self._data_buffer.pop(
                            range(
                                self._index_batch_size
                                if len(self._data_buffer) > self._index_batch_size
                                else len(self._data_buffer)
                            )
                        )
                        self._index.extend(batch_docs)
                        self.logger.debug(f'indexing {len(batch_docs)} docs done...')
            except Exception as e:
                self.logger.error(traceback.format_exc())
                raise e

        self._index_thread = Thread(target=_index_loop, daemon=False)
        self._index_thread.start()

    @requests(on='/update')
    def update(
        self, docs: Optional[DocumentArray] = None, parameters: dict = {}, **kwargs
    ):
        """Update existing documents

        :param docs: the Documents to update
        :param parameters: dictionary with options for updating
        Keys accepted:
            - 'access_paths': traversal paths on docs, e.g. '@r', '@c', '@r,c'
            - 'raise_errors_on_not_found': if True, raise an error if a document is not found. Default is False.
        """

        if not docs:
            return

        access_paths = parameters.get('access_paths', self.index_access_paths)
        raise_errors_on_not_found = parameters.get('raise_errors_on_not_found', False)
        flat_docs = docs[access_paths]
        if len(flat_docs) == 0:
            return

        with self._index_lock:
            if len(self._data_buffer) > 0:
                raise RuntimeError(
                    f'Cannot update documents while the pending documents in the buffer are not indexed yet. '
                    'Please wait for the pending documents to be indexed.'
                )
            for doc in flat_docs:
                try:
                    self._index[doc.id] = doc
                except IndexError:
                    if raise_errors_on_not_found:
                        raise Exception(
                            f'The document (id={doc.id}) cannot be updated as'
                            f'it is not found in the index'
                        )
                    else:
                        self.logger.warning(
                            f'cannot update doc {doc.id} as it does not exist in storage'
                        )

    @requests(on='/delete')
    def delete(self, parameters: dict = {}, **kwargs):
        """Delete existing documents

        Delete entries from the index by id
        :param parameters: parameters to the request
        """

        delete_ids = parameters.get('ids', [])
        if len(delete_ids) == 0:
            return

        with self._index_lock:
            if len(self._data_buffer) > 0:
                raise RuntimeError(
                    f'Cannot delete documents while the pending documents in the buffer are not indexed yet. '
                    'Please wait for the pending documents to be indexed.'
                )

            del self._index[delete_ids]

    @requests(on='/search')
    def search(
        self, docs: Optional[DocumentArray] = None, parameters: dict = {}, **kwargs
    ):
        """Perform a vector similarity search and retrieve Document matches

        Search can be performed with candidate filtering. Filters are a triplet (column,operator,value).
        More than a filter can be applied during search. Therefore, conditions for a filter are specified as a list triplets.
        Each triplet contains:

        - column: Column used to filter.
        - operator: Binary operation between two values. Some supported operators include `['>','<','=','<=','>=']`.
        - value: value used to compare a candidate.

        :param docs: the Documents to search with
        :param parameters: dictionary for parameters for the search operation
        Keys accepted:
            - 'access_paths' (str): traversal paths on docs, e.g. '@r', '@c', '@r,c'
            - 'filter' (dict): the filtering conditions on document tags
            - 'limit' (int): nr of matches to get per Document
        """

        if not docs:
            return

        access_paths = parameters.get('access_paths', self.search_access_paths)
        flat_docs = docs[access_paths]
        match_args = (
            {**self.match_args, **parameters}
            if parameters is not None
            else self.match_args
        )

        with self._index_lock:
            # if len(self._data_buffer) > 0:
            #     raise RuntimeError(
            #         f'Cannot search documents while the pending documents in the buffer are not indexed yet. '
            #         'Please wait for the pending documents to be indexed.'
            #     )

            flat_docs.match(self._index, **match_args)

    @requests(on='/backup')
    def backup(self, parameters: Optional[Dict] = {}, **kwargs):
        """
        Backup data to local or remote.
        Use api of <class 'annlite.index.AnnLite'>

        Keys accepted:
            - 'target' (str): the name of indexer you want to backup as
        """

        target_name = parameters.get('target_name', None)
        token = parameters.get('token', None)
        if target_name:
            target_name = f'{target_name}_{self.runtime_args.shard_id}'
        with self._index_lock:
            if len(self._data_buffer) > 0:
                raise RuntimeError(
                    f'Cannot backup documents while the pending documents in the buffer are not indexed yet. '
                    'Please wait for the pending documents to be indexed.'
                )
            self._index._annlite.backup(target_name, token)

    @requests(on='/restore')
    def restore(self, parameters: Optional[Dict] = {}, **kwargs):
        """
        Restore data from local or remote.
        Use api of <class 'annlite.index.AnnLite'>
        """
        source_name = parameters.get('source_name', None)
        token = parameters.get('token', None)
        if source_name:
            source_name = f'{source_name}_{self.runtime_args.shard_id}'
        self._index._annlite.restore(source_name, token)

    @requests(on='/filter')
    def filter(self, parameters: Dict, **kwargs):
        """
        Query documents from the indexer by the filter `query` object in parameters. The `query` object must follow the
        specifications in the `find` method of `DocumentArray` using annlite: https://docarray.jina.ai/fundamentals/documentarray/find/#filter-with-query-operators
        :param parameters: Dictionary to define the `filter` that you want to use.
        """

        return self._index.find(parameters.get('filter', None))

    @requests(on='/fill_embedding')
    def fill_embedding(self, docs: DocumentArray, **kwargs):
        """
        retrieve embedding of Documents by id
        :param docs: DocumentArray to search with
        """

        for doc in docs:
            doc.embedding = self._index[doc.id].embedding

    @requests(on='/status')
    def status(self, **kwargs) -> DocumentArray:
        """Return the document containing status information about the indexer.

        The status will contain information on the total number of indexed and deleted
        documents, and on the number of (searchable) documents currently in the index.
        """

        status = Document(
            tags={
                'appending_size': len(self._data_buffer),
                'total_docs': len(self._index),
                'index_size': len(self._index),
            }
        )
        return DocumentArray([status])

    def flush(self):
        """Flush all the data in the buffer to the index"""
        while len(self._data_buffer) > 0:
            time.sleep(0.1)

    @requests(on='/clear')
    def clear(self, **kwargs):
        """Clear the index of all entries."""
        self.flush()

        with self._index_lock:
            self._data_buffer = None
            self._index_thread.join()

        self._data_buffer = DocumentArray()
        self._index.clear()

        self._start_index_loop()

    def close(self, **kwargs):
        """Close the index."""
        super().close()

        self.flush()

        # wait for the index thread to finish
        with self._index_lock:
            self._data_buffer = None
            self._index_thread.join()

        # WARNING: the commented code below hangs the close in pytest `pytest tests/test_*.py`
        # But don't know why. It works fine in `pytest tests/test_executor.py` and normal python execution
        del self._index



from typing import List, Optional

import numpy as np
from loguru import logger

from ...math import cdist, top_k
from .base import BaseIndex


class FlatIndex(BaseIndex):
    def __init__(self, *args, **kwargs):
        super(FlatIndex, self).__init__(*args, **kwargs)
        self._data = np.zeros((self.initial_size, self.dim), dtype=self.dtype)

    def search(
        self, x: np.ndarray, limit: int = 10, indices: Optional[np.ndarray] = None
    ):
        _dim = x.shape[-1]
        assert (
            _dim == self.dim
        ), f'the query embedding dimension does not match with index dimension: {_dim} vs {self.dim}'

        x = x.reshape((-1, self.dim))

        data = self._data[: self.size]
        data_ids = np.arange(self.size)

        if indices is not None:
            data = self._data[indices]
            data_ids = data_ids[indices]

        dists = cdist(x, data, metric=self.metric.name.lower())
        dists, idx = top_k(dists, limit, descending=False)

        # TODO: change the shape of return
        dists = dists[0]
        data_ids = data_ids[idx[0]]

        return dists, data_ids

    def add_with_ids(self, x: np.ndarray, ids: List[int]):
        for idx in ids:
            if idx >= self._capacity:
                self._expand_capacity()

        start = self._size
        end = start + len(x)

        self._data[ids, :] = x
        self._size = end

    def _expand_capacity(self):
        new_block = np.zeros((self.expand_step_size, self.dim), dtype=self.dtype)
        self._data = np.concatenate((self._data, new_block), axis=0)

        self._capacity += self.expand_step_size
        logger.debug(
            f'total storage capacity is expanded by {self.expand_step_size}',
        )

    def reset(self, capacity: Optional[int] = None):
        super().reset(capacity=capacity)
        self._data = np.zeros((self.capacity, self.dim), dtype=self.dtype)

    def delete(self, ids: List[int]):
        raise RuntimeError(
            f'the deletion operation is not allowed for {self.__class__.__name__}!'
        )

    def update_with_ids(self, x: np.ndarray, ids: List[int], **kwargs):
        self._data[ids, :] = x



import os
import shutil

import numpy as np
from docarray import Document, DocumentArray


def clean_workspace():
    if os.path.exists('./data'):
        shutil.rmtree('./data')

    if os.path.exists('./workspace'):
        shutil.rmtree('./workspace')


def docs_with_tags(N, D, probs, categories):

    all_docs = []
    start_current = 0
    for k, prob in enumerate(probs):
        n_current = int(N * prob)
        X = np.random.random((n_current, D)).astype(np.float32)

        docs = [
            Document(
                embedding=X[i],
                id=f'{i+start_current}',
                tags={
                    'category': categories[k],
                },
            )
            for i in range(n_current)
        ]
        all_docs.extend(docs)
        start_current += n_current

    return DocumentArray(all_docs)


def _precision(predicted, relevant, eval_at):
    """
    fraction of retrieved documents that are relevant to the query
    """
    if eval_at == 0:
        return 0.0
    predicted_at_k = predicted[:eval_at]
    n_predicted_and_relevant = len(set(predicted_at_k).intersection(set(relevant)))

    return n_predicted_and_relevant / len(predicted)


def _recall(predicted, relevant, eval_at):
    """
    fraction of the relevant documents that are successfully retrieved
    """
    if eval_at == 0:
        return 0.0
    predicted_at_k = predicted[:eval_at]
    n_predicted_and_relevant = len(set(predicted_at_k).intersection(set(relevant)))
    return n_predicted_and_relevant / len(relevant)


def evaluate(predicts, relevants, top_k):
    recall = 0
    precision = 0
    for _predict, _relevant in zip(predicts, relevants):
        _predict = np.array([int(x) for x in _predict])
        recall += _recall(_predict, _relevant, top_k)
        precision += _precision(_predict, _relevant, top_k)

    return recall / len(predicts), precision / len(predicts)



from argparse import ArgumentError

import numpy as np
from scipy.cluster.vq import vq

from annlite import pq_bind

from ...enums import Metric
from ...math import l2_normalize
from ...profile import time_profile
from .base import BaseCodec

# from pqlite.pq_bind import precompute_adc_table, dist_pqcodes_to_codebooks


class PQCodec(BaseCodec):
    """Implementation of Product Quantization (PQ) [Jegou11]_.

    For the indexing phase of database vectors,
    a `D`-dim input vector is divided into `M` `D`/`M`-dim sub-vectors.
    Each sub-vector is quantized into a small integer via `Ks` codewords.

    For the querying phase, given a new `D`-dim query vector, the distance between the query
    and the database PQ-codes are efficiently approximated via Asymmetric Distance.
    All vectors must be np.ndarray with np.float32

    .. [Jegou11] H. Jegou et al., "Product Quantization for Nearest Neighbor Search", IEEE TPAMI 2011

    :param d_vector: the dimensionality of input vectors
    :param n_subvectors: The number of sub-space
    :param n_clusters: The number of codewords for each subspace
            (typically 256, so that each sub-vector is quantized
            into 256 bits pqlite.utils.asymmetric_distance= 1 byte = uint8)
    :param n_init: Number of times K-Means is trained with different centroid seeds. Best result of
                   the `n_init` consecutive runs is selected.
    """

    def __init__(
        self,
        dim: int,
        n_subvectors: int = 8,
        n_clusters: int = 256,
        metric: Metric = Metric.EUCLIDEAN,
        n_init: int = 4,
    ):
        super(PQCodec, self).__init__(require_train=True)
        self.dim = dim
        self.n_subvectors = n_subvectors
        self.n_clusters = n_clusters

        assert (
            dim % n_subvectors == 0
        ), 'input dimension must be dividable by number of sub-space'
        self.d_subvector = dim // n_subvectors

        self.code_dtype = (
            np.uint8
            if n_clusters <= 2**8
            else (np.uint16 if n_clusters <= 2**16 else np.uint32)
        )

        # assert (
        #    metric == Metric.EUCLIDEAN
        # ), f'The distance metric `{metric.name}` is not supported yet!'
        self.metric = metric

        self.normalize_input = False
        if self.metric == Metric.COSINE:
            self.normalize_input = True

        self._codebooks = np.zeros(
            (self.n_subvectors, self.n_clusters, self.d_subvector), dtype=np.float32
        )
        self.kmeans = []
        self.n_init = n_init

    def __hash__(self):
        return hash(
            (
                self.__class__.__name__,
                self.dim,
                self.n_subvectors,
                self.n_clusters,
                self.metric,
                self.code_dtype,
            )
        )

    def fit(self, x: 'np.ndarray', iter: int = 100):
        """Train the K-Means for each cartesian product

        :param x: Training vectors with shape=(N, D)
        :param iter: Number of iterations in Kmeans
        """
        from sklearn.cluster import KMeans

        assert x.dtype == np.float32
        assert x.ndim == 2

        if self.normalize_input:
            x = l2_normalize(x)

        # [m][ks][ds]: m-th subspace, ks-the codeword, ds-th dim
        self._codebooks = np.zeros(
            (self.n_subvectors, self.n_clusters, self.d_subvector), dtype=np.float32
        )
        for m in range(self.n_subvectors):
            kmeans = KMeans(
                n_clusters=self.n_clusters, max_iter=iter, n_init=self.n_init
            )
            self.kmeans.append(kmeans)
            self.kmeans[m].fit(x[:, m * self.d_subvector : (m + 1) * self.d_subvector])
            self._codebooks[m] = self.kmeans[m].cluster_centers_

        self._is_trained = True

    def partial_fit(self, x: 'np.ndarray'):
        """Given a batch of training vectors, update the internal MiniBatchKMeans.
        This method is specially designed to be used when data does not fit in memory.

        :param x: Training vectors with shape=(N, D)
        """
        assert x.ndim == 2

        if self.normalize_input:
            x = l2_normalize(x)

        if len(self.kmeans) > 0:
            for m in range(self.n_subvectors):
                self.kmeans[m].partial_fit(
                    x[:, m * self.d_subvector : (m + 1) * self.d_subvector]
                )
        else:
            from sklearn.cluster import MiniBatchKMeans

            for m in range(self.n_subvectors):
                self.kmeans.append(MiniBatchKMeans(n_clusters=self.n_clusters))

            for m in range(self.n_subvectors):
                self.kmeans[m].partial_fit(
                    x[:, m * self.d_subvector : (m + 1) * self.d_subvector]
                )

    def build_codebook(self):
        """Constructs sub-codebooks from the current parameters of the models in `self.kmeans`
        This step is not necessary if full KMeans is trained used calling `.fit`.
        """

        self._codebooks = np.zeros(
            (self.n_subvectors, self.n_clusters, self.d_subvector), dtype=np.float32
        )

        for m in range(self.n_subvectors):
            self._codebooks[m] = self.kmeans[m].cluster_centers_

        self._is_trained = True

    def encode(self, x: 'np.ndarray'):
        """Encode input vectors into PQ-codes.

        :param x: Input vectors with shape=(N, D) and dtype=np.float32.
        :return: np.ndarray: PQ codes with shape=(N, M) and dtype=self.code_dtype
        """
        assert x.dtype == np.float32
        assert x.ndim == 2
        N, D = x.shape
        assert (
            D == self.d_subvector * self.n_subvectors
        ), 'input dimension must be Ds * M'

        # codes[n][m] : code of n-th vec, m-th subspace
        codes = np.empty((N, self.n_subvectors), dtype=self.code_dtype)
        for m in range(self.n_subvectors):
            sub_vecs = x[:, m * self.d_subvector : (m + 1) * self.d_subvector]
            codes[:, m], _ = vq(sub_vecs, self.codebooks[m])

        return codes

    def decode(self, codes: 'np.ndarray'):
        """Given PQ-codes, reconstruct original D-dimensional vectors
        approximately by fetching the codewords.

        :param codes: PQ-cdoes with shape=(N, M) and dtype=self.code_dtype.
            Each row is a PQ-code
        :return: Reconstructed vectors with shape=(N, D) and dtype=np.float32
        """
        assert codes.ndim == 2
        N, M = codes.shape
        assert M == self.n_subvectors
        assert codes.dtype == self.code_dtype

        vecs = np.empty((N, self.d_subvector * self.n_subvectors), dtype=np.float32)
        for m in range(self.n_subvectors):
            vecs[:, m * self.d_subvector : (m + 1) * self.d_subvector] = self.codebooks[
                m
            ][codes[:, m], :]

        return vecs

    def precompute_adc(self, query: object) -> object:
        """Compute a distance table for a query vector.
        The distances are computed by comparing each sub-vector of the query
        to the codewords for each sub-subspace.
        `dtable[m][ks]` contains the squared Euclidean distance between
        the `m`-th sub-vector of the query and the `ks`-th codeword
        for the `m`-th sub-space (`self.codewords[m][ks]`).

        :param query: Input vector with shape=(D, ) and dtype=np.float32
        :return: Distance table. which contains dtable with shape=(M, Ks)
            and dtype=np.float32
        """
        assert query.dtype == np.float32
        assert query.ndim == 1, 'input must be a single vector'

        # dtable[m] : distance between m-th subvec and m-th codewords (m-th subspace)
        # dtable[m][ks] : distance between m-th subvec and ks-th codeword of m-th codewords

        # Warning: the following line produces `ValueError: buffer source array is read-only`
        # if no `const` is used in the cython implementation using a memoryview
        dtable = pq_bind.precompute_adc_table(
            query, self.d_subvector, self.n_clusters, self.codebooks
        )

        return DistanceTable(dtable)

    @property
    def codebooks(self):
        return self._codebooks

    # trained pq interface ----------------
    def get_codebook(self) -> 'np.ndarray':
        """Return the codebook parameters.

        Expect a 3-dimensional matrix is returned,
        with shape (`n_subvectors`, `n_clusters`, `d_subvector`) and dtype float32
        """
        return np.ascontiguousarray(self.codebooks, dtype='float32')

    def get_subspace_splitting(self):
        """Return subspace splitting setting

        :return: tuple of (`n_subvectors`, `n_clusters`, `d_subvector`)
        """
        return (self.n_subvectors, self.n_clusters, self.d_subvector)

    # def get_dist_mat(self, x: np.ndarray):
    #     """Return the distance tables in form of matrix for multiple queries

    #     :param query: shape('N', 'D'),

    #     :return: ndarray with shape('N', `n_subvectors`, `n_clusters`)

    #     .. note::
    #         _description_
    #     """
    #     assert x.dtype == np.float32
    #     assert x.ndim == 2
    #     N, D = x.shape
    #     assert (
    #         D == self.d_subvector * self.n_subvectors
    #     ), 'input dimension must be Ds * M'
    #     if self.normalize_input:
    #         x = l2_normalize(x)

    #     x = x.reshape(
    #         N,
    #         self.n_subvectors,
    #         1,
    #         self.d_subvector,
    #     )
    #     if self.metric == Metric.EUCLIDEAN:
    #         # (1, n_subvectors, n_clusters, d_subvector)
    #         codebook = self.codebooks[np.newaxis, ...]

    #         # broadcast to (N, n_subvectors, n_clusters, d_subvector)
    #         dist_vector = (x - codebook) ** 2

    #         # reduce to (N, n_subvectors, n_clusters)
    #         dist_mat = np.sum(dist_vector, axis=3)
    #     elif self.metric in [Metric.INNER_PRODUCT, Metric.COSINE]:
    #         # (1, n_subvectors, n_clusters, d_subvector)
    #         codebook = self.codebooks[np.newaxis, ...]

    #         # broadcast to (N, n_subvectors, n_clusters, d_subvector)
    #         dist_vector = x * codebook

    #         # reduce to (N, n_subvectors, n_clusters)
    #         dist_mat = 1 / self.n_clusters - np.sum(dist_vector, axis=3)
    #     else:
    #         raise ArgumentError(f'Unable support metrics {self.metric}')
    #     return np.ascontiguousarray(dist_mat, dtype='float32')

    def get_dist_mat(self, x: np.ndarray):
        """Return the distance tables in form of matrix for multiple queries

        :param query: shape('N', 'D'),

        :return: ndarray with shape('N', `n_subvectors`, `n_clusters`)

        .. note::
            _description_
        """
        assert x.dtype == np.float32
        assert x.ndim == 2
        N, D = x.shape
        assert (
            D == self.d_subvector * self.n_subvectors
        ), 'input dimension must be Ds * M'
        if self.normalize_input:
            x = l2_normalize(x)

        if self.metric == Metric.EUCLIDEAN:
            dist_mat = pq_bind.batch_precompute_adc_table(
                x, self.d_subvector, self.n_clusters, self.codebooks
            )
        elif self.metric in [Metric.INNER_PRODUCT, Metric.COSINE]:
            dist_mat = 1 / self.n_clusters - np.array(
                pq_bind.batch_precompute_adc_table_ip(
                    x, self.d_subvector, self.n_clusters, self.codebooks
                ),
                dtype='float32',
            )
        else:
            raise ArgumentError(f'Unable support metrics {self.metric}')
        return np.ascontiguousarray(dist_mat, dtype='float32')

    # -------------------------------------


class DistanceTable(object):
    """Distance table from query to codeworkds.
    Given a query vector, a PQ/OPQ instance compute this DistanceTable class
    using :func:`PQ.dtable` or :func:`OPQ.dtable`.
    The Asymmetric Distance from query to each database codes can be computed
    by :func:`DistanceTable.adist`.
    Args:
        dtable (np.ndarray): Distance table with shape=(M, Ks) and dtype=np.float32
            computed by :func:`PQ.dtable` or :func:`OPQ.dtable`
    Attributes:
        dtable (np.ndarray): Distance table with shape=(M, Ks) and dtype=np.float32.
            Note that dtable[m][ks] contains the squared Euclidean distance between
            (1) m-th sub-vector of query and (2) ks-th codeword for m-th subspace.
    """

    def __init__(self, dtable: 'np.ndarray'):

        assert dtable.ndim == 2
        self.dtable = dtable

    def adist(self, codes):
        """Given PQ-codes, compute Asymmetric Distances between the query (self.dtable)
        and the PQ-codes.
        Args:
            codes (np.ndarray): PQ codes with shape=(N, M) and
                dtype=pq.code_dtype where pq is a pq instance that creates the codes
        Returns:
            np.ndarray: Asymmetric Distances with shape=(N, ) and dtype=np.float32
        """

        assert codes.ndim == 2
        dists = pq_bind.dist_pqcodes_to_codebooks(self.dtable, codes)

        # The above line is equivalent to the followings:
        # dists = np.zeros((N, )).astype(np.float32)
        # for n in range(N):
        #     for m in range(M):
        #         dists[n] += self.dtable[m][codes[n][m]]
        return dists



import sys

import numpy as np
from loguru import logger


def setup_logging(debug: bool):
    """
    Setup the log formatter for AnnLite.
    """

    log_level = 'INFO'
    if debug:
        log_level = 'DEBUG'

    logger.remove()
    logger.add(
        sys.stdout,
        colorize=True,
        level=log_level,
    )


def str2dtype(dtype_str: str):
    if dtype_str in ['double', 'float64']:
        dtype = np.float64
    elif dtype_str in ['half', 'float16']:
        dtype = np.float16
    elif dtype_str in ['float', 'float32']:
        dtype = np.float32
    elif dtype_str in ['bfloat16']:
        dtype = np.bfloat16
    elif dtype_str in ['long', 'int64']:
        dtype = np.int64
    elif dtype_str in ['int', 'int32']:
        dtype = np.int32
    elif dtype_str in ['int16']:
        dtype = np.int16
    elif dtype_str in ['int8']:
        dtype = np.int8
    elif dtype_str in ['uint8']:
        dtype = np.uint8
    elif dtype_str in ['bool']:
        dtype = np.bool
    else:
        raise TypeError(f'Unrecognized dtype string: {dtype_str}')
    return dtype



import abc
from typing import TYPE_CHECKING, List, Optional

if TYPE_CHECKING:
    import numpy as np

from ..enums import ExpandMode


class Storage(abc.ABC):
    def __init__(
        self,
        initial_size: Optional[int] = None,
        expand_step_size: int = 10240,
        expand_mode: ExpandMode = ExpandMode.ADAPTIVE,
    ):
        if initial_size is None:
            initial_size = expand_step_size
        assert initial_size >= 0
        assert expand_step_size > 0

        self.initial_size = initial_size
        self.expand_step_size = expand_step_size
        self.expand_mode = expand_mode

    @property
    @abc.abstractmethod
    def capacity(self) -> int:
        ...

    @property
    @abc.abstractmethod
    def size(self):
        ...

    @abc.abstractmethod
    def clean(self):
        ...

    @abc.abstractmethod
    def add(
        self,
        data: 'np.ndarray',
        cells: 'np.ndarray',
        ids: List[str],
        doc_tags: Optional[List[dict]] = None,
    ):
        ...

    @abc.abstractmethod
    def delete(self, ids: List[str]):
        ...

    @abc.abstractmethod
    def update(
        self,
        data: 'np.ndarray',
        cells: 'np.ndarray',
        ids: List[str],
        doc_tags: Optional[List[dict]] = None,
    ):
        ...



import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple

import numpy as np
from docarray import Document, DocumentArray
from loguru import logger

if TYPE_CHECKING:
    from .core.codec.pq import PQCodec
    from .core.codec.projector import ProjectorCodec

from .core.index.hnsw import HnswIndex
from .enums import Metric
from .storage.base import ExpandMode
from .storage.kv import DocStorage
from .storage.table import CellTable, MetaTable

VALID_FILTERABLE_DATA_TYPES = [int, str, float]


class CellContainer:
    def __init__(
        self,
        n_dim: int,
        metric: Metric = Metric.COSINE,
        n_cells: int = 1,
        projector_codec: Optional['ProjectorCodec'] = None,
        pq_codec: Optional['PQCodec'] = None,
        initial_size: Optional[int] = None,
        expand_step_size: int = 50000,
        expand_mode: 'ExpandMode' = ExpandMode.STEP,
        filterable_attrs: Optional[Dict] = None,
        serialize_config: Optional[Dict] = None,
        data_path: 'Path' = Path('./data'),
        **kwargs,
    ):
        self.n_dim = n_dim
        self.metric = metric
        self.n_cells = n_cells
        self.n_components = projector_codec.n_components if projector_codec else None
        self.data_path = data_path
        self.serialize_config = serialize_config

        self._pq_codec = pq_codec
        self._projector_codec = projector_codec

        self._vec_indexes = [
            HnswIndex(
                dim=self.n_components or n_dim,
                metric=metric,
                initial_size=initial_size,
                expand_step_size=expand_step_size,
                expand_mode=expand_mode,
                pq_codec=pq_codec,
                **kwargs,
            )
            for _ in range(n_cells)
        ]

        self._doc_stores = [
            DocStorage(
                data_path / f'cell_{_}',
                serialize_config=serialize_config or {},
                lock=True,
            )
            for _ in range(n_cells)
        ]

        columns = []
        if filterable_attrs:
            for attr_name, attr_type in filterable_attrs.items():
                if isinstance(attr_type, str):
                    attr_type = eval(attr_type)

                if attr_type not in VALID_FILTERABLE_DATA_TYPES:
                    raise ValueError(
                        f'Invalid filterable attribute type `{attr_type}` for attribute `{attr_name}`. '
                    )
                columns.append((attr_name, attr_type))

        self._cell_tables = [
            CellTable(f'table_{c}', columns=columns) for c in range(n_cells)
        ]

        self._meta_table = MetaTable('metas', data_path=data_path, in_memory=False)

    def ivf_search(
        self,
        x: 'np.ndarray',
        cells: 'np.ndarray',
        where_clause: str = '',
        where_params: Tuple = (),
        limit: int = 10,
    ):
        dists = []

        doc_idx = []
        cell_ids = []
        count = 0
        for cell_id in cells:
            cell_table = self.cell_table(cell_id)
            cell_size = cell_table.count()
            if cell_size == 0:
                continue

            indices = None
            if where_clause or (cell_table.deleted_count() > 0):
                indices = cell_table.query(
                    where_clause=where_clause, where_params=where_params
                )

                if len(indices) == 0:
                    continue

                indices = np.array(indices, dtype=np.int64)

            _dists, _doc_idx = self.vec_index(cell_id).search(
                x, limit=min(limit, cell_size), indices=indices
            )

            if count >= limit and _dists[0] > dists[-1][-1]:
                continue

            dists.append(_dists)
            doc_idx.append(_doc_idx)
            cell_ids.extend([cell_id] * len(_dists))
            count += len(_dists)

        cell_ids = np.array(cell_ids, dtype=np.int64)
        if len(dists) != 0:
            dists = np.hstack(dists)
            doc_idx = np.hstack(doc_idx)

            indices = dists.argsort(axis=0)[:limit]
            dists = dists[indices]
            cell_ids = cell_ids[indices]
            doc_idx = doc_idx[indices]

        doc_ids = []
        for cell_id, offset in zip(cell_ids, doc_idx):
            doc_id = self.cell_table(cell_id).get_docid_by_offset(offset)
            doc_ids.append(doc_id)
        return dists, doc_ids, cell_ids

    def filter_cells(
        self,
        cells: 'np.ndarray',
        where_clause: str = '',
        where_params: Tuple = (),
        limit: int = -1,
        offset: int = 0,
        order_by: Optional[str] = None,
        ascending: bool = True,
        include_metadata: bool = False,
    ):
        result = DocumentArray()
        if len(cells) > 1 and offset > 0:
            raise ValueError('Offset is not supported for multiple cells')

        for cell_id in cells:
            cell_table = self.cell_table(cell_id)
            cell_size = cell_table.count()
            if cell_size == 0:
                continue

            indices = cell_table.query(
                where_clause=where_clause,
                where_params=where_params,
                order_by=order_by,
                limit=limit,
                offset=offset,
                ascending=ascending,
            )

            if len(indices) == 0:
                continue

            for offset in indices:
                doc_id = self.cell_table(cell_id).get_docid_by_offset(offset)
                doc = Document(id=doc_id)
                if include_metadata or (len(cells) > 1 and order_by):
                    doc = self.doc_store(cell_id).get([doc_id])[0]

                result.append(doc)

            if not order_by and len(result) >= limit > 0:
                break

        # reordering the results from multiple cells
        if order_by and len(cells) > 1:
            result = sorted(
                result, key=lambda d: d.tags.get(order_by), reverse=not ascending
            )
            if limit > 0:
                result = result[:limit]
            result = DocumentArray(result)

        return result

    def search_cells(
        self,
        query: 'np.ndarray',
        cells: 'np.ndarray',
        where_clause: str = '',
        where_params: Tuple = (),
        limit: int = 10,
        include_metadata: bool = False,
    ):
        if self._projector_codec:
            query = self._projector_codec.encode(query)

        topk_dists, topk_docs = [], []
        for x, cell_idx in zip(query, cells):
            # x.shape = (self.n_dim,)
            dists, doc_ids, cells = self.ivf_search(
                x,
                cells=cell_idx,
                where_clause=where_clause,
                where_params=where_params,
                limit=limit,
            )

            topk_dists.append(dists)
            match_docs = DocumentArray()
            for dist, doc_id, cell_id in zip(dists, doc_ids, cells):
                doc = Document(id=doc_id)
                if include_metadata:
                    doc = self.doc_store(cell_id).get([doc_id])[0]

                doc.scores[self.metric.name.lower()].value = dist
                match_docs.append(doc)
            topk_docs.append(match_docs)

        return topk_dists, topk_docs

    def _search_cells(
        self,
        query: 'np.ndarray',
        cells: 'np.ndarray',
        where_clause: str = '',
        where_params: Tuple = (),
        limit: int = 10,
    ):
        if self._projector_codec:
            query = self._projector_codec.encode(query)

        topk_dists, topk_ids = [], []
        for x, cell_idx in zip(query, cells):
            dists, ids, cells = self.ivf_search(
                x,
                cells=cell_idx,
                where_clause=where_clause,
                where_params=where_params,
                limit=limit,
            )
            topk_dists.append(dists)
            topk_ids.append(ids)

        return topk_dists, [np.array(ids, dtype=int) for ids in topk_ids]

    def insert(
        self,
        data: 'np.ndarray',
        cells: 'np.ndarray',
        docs: 'DocumentArray',
        only_index: bool = False,
    ):
        assert len(docs) == len(data)

        if self._projector_codec:
            data = self._projector_codec.encode(data)

        unique_cells, unique_cell_counts = np.unique(cells, return_counts=True)

        if len(unique_cells) == 1:
            cell_id = unique_cells[0]

            offsets = self.cell_table(cell_id).insert(docs)
            offsets = np.array(offsets, dtype=np.int64)

            self.vec_index(cell_id).add_with_ids(data, offsets)

            if not only_index:
                self.doc_store(cell_id).insert(docs)
                self._meta_table.bulk_add_address([d.id for d in docs], cells, offsets)
        else:
            for cell_id, cell_count in zip(unique_cells, unique_cell_counts):
                # TODO: Jina should allow boolean filtering in docarray to avoid this
                # and simply use cells == cell_index
                indices = np.where(cells == cell_id)[0]
                cell_docs = docs[indices.tolist()]

                cell_offsets = self.cell_table(cell_id).insert(cell_docs)
                cell_offsets = np.array(cell_offsets, dtype=np.int64)

                cell_data = data[indices, :]

                self.vec_index(cell_id).add_with_ids(cell_data, cell_offsets)

                if not only_index:
                    self.doc_store(cell_id).insert(cell_docs)
                    self._meta_table.bulk_add_address(
                        [d.id for d in cell_docs], [cell_id] * cell_count, cell_offsets
                    )
        logger.debug(f'{len(docs)} new docs added')

    def _add_vecs(self, data: 'np.ndarray', cells: 'np.ndarray', offsets: 'np.ndarray'):
        assert data.shape[0] == cells.shape[0]
        assert data.shape[1] == self.n_dim

        unique_cells, _ = np.unique(cells, return_counts=True)

        for cell_id in unique_cells:
            indices = cells == cell_id
            x = data[indices, :]
            ids = offsets[indices]

            self.vec_index(cell_id).add_with_ids(x, ids)

    def update(
        self,
        data: 'np.ndarray',
        cells: 'np.ndarray',
        docs: 'DocumentArray',
        insert_if_not_found: bool = True,
        raise_errors_on_not_found: bool = False,
    ):
        update_success = 0

        new_data = []
        new_cells = []
        new_docs = []

        for (
            x,
            doc,
            cell_id,
        ) in zip(data, docs, cells):
            _cell_id, _offset = self._meta_table.get_address(doc.id)
            if cell_id == _cell_id:
                self.vec_index(cell_id).add_with_ids(x.reshape(1, -1), [_offset])
                self.cell_table(cell_id).undo_delete_by_offset(_offset)
                self.doc_store(cell_id).update([doc])
                self.meta_table.add_address(doc.id, cell_id, _offset)
                update_success += 1
            elif _cell_id is None:
                if raise_errors_on_not_found and not insert_if_not_found:
                    raise Exception(
                        f'The document (id={doc.id}) cannot be updated as'
                        f'it is not found in the index'
                    )
                elif not (raise_errors_on_not_found or insert_if_not_found):
                    warnings.warn(
                        f'The document (id={doc.id}) cannot be updated as '
                        f'it is not found in the index',
                        RuntimeWarning,
                    )
                elif insert_if_not_found:
                    new_data.append(x)
                    new_cells.append(cell_id)
                    new_docs.append(doc)
                    update_success += 1
                else:
                    continue
            else:
                # DELETE and INSERT
                self.vec_index(_cell_id).delete(_offset)
                self.cell_table(_cell_id).delete_by_offset(_offset)
                self.doc_store(_cell_id).delete([doc.id])

                new_data.append(x)
                new_cells.append(cell_id)
                new_docs.append(doc)
                update_success += 1

        if len(new_data) > 0:
            new_data = np.stack(new_data)
            new_cells = np.array(new_cells, dtype=np.int64)

            self.insert(new_data, new_cells, new_docs)

        logger.debug(
            f'total items for updating: {len(docs)}, ' f'success: {update_success}'
        )

    def delete(
        self,
        ids: List[str],
        raise_errors_on_not_found: bool = False,
    ):
        delete_success = 0

        for doc_id in ids:
            cell_id, offset = self._meta_table.get_address(doc_id)
            print(f'{doc_id} {cell_id} {offset}')
            if cell_id is not None:
                self.vec_index(cell_id).delete([offset])
                self.cell_table(cell_id).delete_by_offset(offset)
                self.doc_store(cell_id).delete([doc_id])
                self.meta_table.delete_address(doc_id)
                delete_success += 1
            else:
                if raise_errors_on_not_found:
                    raise Exception(
                        f'The document (id={doc_id}) cannot be updated as'
                        f'it is not found in the index'
                    )
                else:
                    continue

        logger.debug(
            f'total items for updating: {len(ids)}, ' f'success: {delete_success}'
        )

    def _rebuild_database(self):
        """rebuild doc_store and meta_table after annlite download databse from hubble"""

        self._doc_stores = [
            DocStorage(
                self.data_path / f'cell_{_}',
                serialize_config=self.serialize_config or {},
                lock=True,
            )
            for _ in range(self.n_cells)
        ]
        self._meta_table = MetaTable('metas', data_path=self.data_path, in_memory=False)

    def _get_doc_by_id(self, doc_id: str):
        cell_id = 0
        if self.n_cells > 1:
            cell_id, _ = self._meta_table.get_address(doc_id)

        da = self.doc_store(cell_id).get([doc_id])
        return da[0] if len(da) > 0 else None

    def documents_generator(self, cell_id: int, batch_size: int = 1000):
        for docs in self.doc_store(cell_id).batched_iterator(batch_size=batch_size):
            yield docs

    @property
    def cell_tables(self):
        return self._cell_tables

    @property
    def cell_indexes(self):
        return self._vec_indexes

    def cell_table(self, cell_id: int):
        return self._cell_tables[cell_id]

    def doc_store(self, cell_id: int):
        return self._doc_stores[cell_id]

    def vec_index(self, cell_id: int):
        return self._vec_indexes[cell_id]

    @property
    def meta_table(self):
        return self._meta_table

    @property
    def total_docs(self):
        return sum([store.size for store in self._doc_stores])

    @property
    def index_size(self):
        return sum([table.size for table in self._cell_tables])



import hashlib
import logging
import os
import platform
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Optional, Union

import numpy as np
from docarray.math.ndarray import to_numpy_array
from loguru import logger

if TYPE_CHECKING:
    from docarray import DocumentArray

from .container import CellContainer
from .core import PQCodec, ProjectorCodec, VQCodec
from .enums import Metric
from .filter import Filter
from .helper import setup_logging
from .math import cdist, top_k

MAX_TRAINING_DATA_SIZE = 10240


class AnnLite(CellContainer):
    """:class:`AnnLite` is an approximate nearest neighbor search library.

    To create a :class:`AnnLite` object, simply:

        .. highlight:: python
        .. code-block:: python
            ann = AnnLite(256, metric='cosine')

    :param n_dim: dimensionality of input vectors. there are 2 constraints on dim:
            (1) it needs to be divisible by n_subvectors; (2) it needs to be a multiple of 4.*
    :param metric: distance metric type, can be 'euclidean', 'inner_product', or 'cosine'.
    :param n_subvectors: number of sub-quantizers, essentially this is the byte size of
            each quantized vector, default is None.
    :param n_cells:  number of coarse quantizer clusters, default is 1.
    :param n_probe: number of cells to search for each query, default is 16.
    :param n_components: number of components to keep.
    :param initial_size: initial capacity assigned to each voronoi cell of coarse quantizer.
            ``n_cells * initial_size`` is the number of vectors that can be stored initially.
            if any cell has reached its capacity, that cell will be automatically expanded.
            If you need to add vectors frequently, a larger value for init_size is recommended.
    :param columns: the columns to be indexed for fast filtering, default is None.
    :param filterable_attrs: a dict of attributes to be indexed for fast filtering, default is None.
            The key is the attribute name, and the value is the attribute type. And it only works when ``columns`` is None.
    :param data_path: path to the directory where the data is stored.
    :param create_if_missing: if False, do not create the directory path if it is missing.
    :param read_only: if True, the index is not writable.
    :param verbose: if True, will print the debug logging info.

    .. note::
        Remember that the shape of any tensor that contains data points has to be `[n_data, dim]`.
    """

    def __init__(
        self,
        n_dim: int,
        metric: Union[str, Metric] = 'cosine',
        n_cells: int = 1,
        n_subvectors: Optional[int] = None,
        n_clusters: Optional[int] = 256,
        n_probe: int = 16,
        n_components: Optional[int] = None,
        initial_size: Optional[int] = None,
        expand_step_size: int = 10240,
        columns: Optional[Union[Dict, List]] = None,
        filterable_attrs: Optional[Dict] = None,
        data_path: Union[Path, str] = Path('./data'),
        create_if_missing: bool = True,
        read_only: bool = False,
        verbose: bool = False,
        **kwargs,
    ):
        setup_logging(verbose)

        if 'dim' in kwargs:
            warnings.warn(
                'The argument `dim` will be deprecated, please use `n_dim` instead.'
            )
            n_dim = kwargs['dim']

        if n_subvectors:
            assert (
                n_dim % n_subvectors == 0
            ), '"n_dim" needs to be divisible by "n_subvectors"'
        self.n_dim = n_dim
        self.n_components = n_components
        self.n_subvectors = n_subvectors
        self.n_clusters = n_clusters
        self.n_probe = max(n_probe, n_cells)
        self.n_cells = n_cells
        self.size_limit = 2048

        if isinstance(metric, str):
            metric = Metric.from_string(metric)
        self.metric = metric

        self._use_smart_probing = True

        self.read_only = read_only

        data_path = Path(data_path)
        if create_if_missing:
            data_path.mkdir(parents=True, exist_ok=True)
        self.data_path = data_path

        self._projector_codec = None
        if self._projector_codec_path.exists():
            logger.info(
                f'Load pre-trained projector codec (n_components={self.n_components}) from {self.model_path}'
            )
            self._projector_codec = ProjectorCodec.load(self._projector_codec_path)
        elif n_components:
            logger.info(
                f'Initialize Projector codec (n_components={self.n_components})'
            )
            self._projector_codec = ProjectorCodec(
                n_dim, n_components=self.n_components
            )

        self._vq_codec = None
        if self._vq_codec_path.exists():
            logger.info(
                f'Load trained VQ codec (K={self.n_cells}) from {self.model_path}'
            )
            self._vq_codec = VQCodec.load(self._vq_codec_path)
        elif n_cells > 1:
            logger.info(f'Initialize VQ codec (K={self.n_cells})')
            self._vq_codec = VQCodec(self.n_cells, metric=self.metric)

        self._pq_codec = None
        if self._pq_codec_path.exists():
            logger.info(
                f'Load trained PQ codec (n_subvectors={self.n_subvectors}) from {self.model_path}'
            )
            self._pq_codec = PQCodec.load(self._pq_codec_path)
        elif n_subvectors:
            logger.info(f'Initialize PQ codec (n_subvectors={self.n_subvectors})')
            self._pq_codec = PQCodec(
                dim=n_dim
                if not self._projector_codec
                else self._projector_codec.n_components,
                n_subvectors=self.n_subvectors,
                n_clusters=self.n_clusters,
                metric=self.metric,
            )

        if columns is not None:
            if filterable_attrs:
                logger.warning('`filterable_attrs` will be overwritten by `columns`.')

            filterable_attrs = {}
            for n, t in columns.items() if isinstance(columns, dict) else columns:
                filterable_attrs[n] = t

        super(AnnLite, self).__init__(
            n_dim,
            metric=metric,
            projector_codec=self._projector_codec,
            pq_codec=self._pq_codec,
            n_cells=n_cells,
            initial_size=initial_size,
            expand_step_size=expand_step_size,
            filterable_attrs=filterable_attrs,
            data_path=data_path,
            **kwargs,
        )

        if not self.is_trained and self.total_docs > 0:
            # train the index from scratch based on the data in the data_path
            logger.info(f'Train the index by reading data from {self.data_path}')
            total_size = 0
            # TODO: add a progress bar
            for docs in self.documents_generator(0, batch_size=1024):
                x = to_numpy_array(docs.embeddings)
                total_size += x.shape[0]
                self.partial_train(x, auto_save=True, force_train=True)
                if total_size >= MAX_TRAINING_DATA_SIZE:
                    break
            logger.info(f'Total training data size: {total_size}')

        if self.total_docs > 0:
            self.restore()

    def _sanity_check(self, x: 'np.ndarray'):
        assert x.ndim == 2, 'inputs must be a 2D array'
        assert (
            x.shape[1] == self.n_dim
        ), f'inputs must have the same dimension as the index , got {x.shape[1]}, expected {self.n_dim}'

        return x.shape

    def train(self, x: 'np.ndarray', auto_save: bool = True, force_train: bool = False):
        """Train the index with the given data.

        :param x: the ndarray data for training.
        :param auto_save: if False, will not dump the trained model to ``model_path``.
        :param force_train: if True, enforce to retrain the model, and overwrite the model if ``auto_save=True``.
        """
        n_data, _ = self._sanity_check(x)

        if self.is_trained and not force_train:
            logger.warning(
                'The indexer has been trained or is not trainable. Please use ``force_train=True`` to retrain.'
            )
            return

        if self._projector_codec:
            logger.info(
                f'Start training Projector codec (n_components={self.n_components}) with {n_data} data...'
            )
            self._projector_codec.fit(x)

        if self._vq_codec:
            logger.info(
                f'Start training VQ codec (K={self.n_cells}) with {n_data} data...'
            )
            self._vq_codec.fit(x)

        if self._pq_codec:
            logger.info(
                f'Start training PQ codec (n_subvectors={self.n_subvectors}) with {n_data} data...'
            )
            self._pq_codec.fit(x)

        logger.info(f'The annlite is successfully trained!')

        if auto_save:
            self.dump_model()

    def partial_train(
        self, x: np.ndarray, auto_save: bool = True, force_train: bool = False
    ):
        """Partially train the index with the given data.

        :param x: the ndarray data for training.
        :param auto_save: if False, will not dump the trained model to ``model_path``.
        :param force_train: if True, enforce to retrain the model, and overwrite the model if ``auto_save=True``.

        """
        n_data, _ = self._sanity_check(x)

        if self.is_trained and not force_train:
            logger.warning(
                'The annlite has been trained or is not trainable. Please use ``force_train=True`` to retrain.'
            )
            return

        if self._projector_codec:
            logging.info(
                f'Partial training Projector codec (n_components={self.n_components}) with {n_data} data...'
            )
            self._projector_codec.partial_fit(x)

        if self._vq_codec:
            logger.info(
                f'Partial training VQ codec (K={self.n_cells}) with {n_data} data...'
            )
            self._vq_codec.partial_fit(x)

        if self._pq_codec:
            logger.info(
                f'Partial training PQ codec (n_subvectors={self.n_subvectors}) with {n_data} data...'
            )
            self._pq_codec.partial_fit(x)

        if auto_save:
            self.dump_model()

    def index(self, docs: 'DocumentArray', **kwargs):
        """Add the documents to the index.

        :param docs: the document array to be indexed.
        """

        if self.read_only:
            logger.error('The indexer is readonly, cannot add new documents')
            return

        if not self.is_trained:
            raise RuntimeError(f'The indexer is not trained, cannot add new documents')

        x = to_numpy_array(docs.embeddings)
        n_data, _ = self._sanity_check(x)

        assigned_cells = (
            self._vq_codec.encode(x)
            if self._vq_codec
            else np.zeros(n_data, dtype=np.int64)
        )
        return super(AnnLite, self).insert(x, assigned_cells, docs)

    def update(
        self,
        docs: 'DocumentArray',
        raise_errors_on_not_found: bool = False,
        insert_if_not_found: bool = True,
        **kwargs,
    ):
        """Update the documents in the index.

        :param insert_if_not_found: whether to raise error when updated id is not found.
        :param raise_errors_on_not_found: whether to raise exception when id not found.
        :param docs: the document array to be updated.
        """
        if self.read_only:
            logger.error('The indexer is readonly, cannot update documents')
            return

        if not self.is_trained:
            raise RuntimeError(f'The indexer is not trained, cannot add new documents')

        x = to_numpy_array(docs.embeddings)
        n_data, _ = self._sanity_check(x)

        assigned_cells = (
            self._vq_codec.encode(x)
            if self._vq_codec
            else np.zeros(n_data, dtype=np.int64)
        )

        return super(AnnLite, self).update(
            x,
            assigned_cells,
            docs,
            raise_errors_on_not_found=raise_errors_on_not_found,
            insert_if_not_found=insert_if_not_found,
        )

    def search(
        self,
        docs: 'DocumentArray',
        filter: Optional[dict] = None,
        limit: int = 10,
        include_metadata: bool = True,
        **kwargs,
    ):
        """Search the index, and attach matches to the query Documents in `docs`

        :param docs: the document array to be searched.
        :param filter: the filter to be applied to the search.
        :param limit: the number of results to get for each query document in search
        :param include_metadata: whether to return document metadata in response.
        """
        if not self.is_trained:
            raise RuntimeError(f'The indexer is not trained, cannot add new documents')

        query_np = to_numpy_array(docs.embeddings)

        match_dists, match_docs = self.search_by_vectors(
            query_np, filter=filter, limit=limit, include_metadata=include_metadata
        )

        for doc, matches in zip(docs, match_docs):
            doc.matches = matches

    def search_by_vectors(
        self,
        query_np: 'np.ndarray',
        filter: Optional[dict] = None,
        limit: int = 10,
        include_metadata: bool = True,
    ):
        """Search the index by vectors, and return the matches.

        :param query_np: the query vectors.
        :param filter: the filter to be applied to the search.
        :param limit: the number of results to get for each query document in search
        :param include_metadata: whether to return document metadata in response.
        """

        cells = self._cell_selection(query_np, limit)
        where_clause, where_params = Filter(filter or {}).parse_where_clause()

        match_dists, match_docs = self.search_cells(
            query=query_np,
            cells=cells,
            where_clause=where_clause,
            where_params=where_params,
            limit=limit,
            include_metadata=include_metadata,
        )
        return match_dists, match_docs

    def filter(
        self,
        filter: Dict,
        limit: int = 10,
        offset: int = 0,
        order_by: Optional[str] = None,
        ascending: bool = True,
        include_metadata: bool = True,
    ):
        """Find the documents by the filter.

        :param filter: the filter to be applied to the search.
        :param limit: the number of results.
        :param offset: the offset of the results.
        :param order_by: the field to order the results.
        :param ascending: whether to order the results in ascending order.
        :param include_metadata: whether to return document metadata in response.
        """
        cells = [x for x in range(self.n_cells)]
        where_clause, where_params = Filter(filter or {}).parse_where_clause()

        match_docs = self.filter_cells(
            cells=cells,
            where_clause=where_clause,
            where_params=where_params,
            limit=limit,
            offset=offset,
            order_by=order_by,
            ascending=ascending,
            include_metadata=include_metadata,
        )
        if limit > 0:
            return match_docs[:limit]
        return match_docs

    def get_doc_by_id(self, doc_id: str):
        """Get the document by id.

        :param doc_id: the document id.
        """

        return self._get_doc_by_id(doc_id)

    def get_docs(
        self,
        filter: Optional[dict] = None,
        limit: int = 10,
        offset: int = 0,
        order_by: Optional[str] = None,
        ascending: bool = True,
    ):
        """Get the documents.

        :param filter: the filter to be applied to the search.
        :param limit: the number of results.
        :param offset: the offset of the results.
        :param order_by: the field to order the results.
        :param ascending: whether to order the results in ascending order. It only works when `order_by` is specified.
        """

        return self.filter(
            filter=filter,
            limit=limit,
            offset=offset,
            order_by=order_by,
            ascending=ascending,
            include_metadata=True,
        )

    def _cell_selection(self, query_np, limit):
        n_data, _ = self._sanity_check(query_np)

        if self._vq_codec:
            dists = cdist(
                query_np, self._vq_codec.codebook, metric=self.metric.name.lower()
            )
            dists, cells = top_k(dists, k=self.n_probe)
        else:
            cells = np.zeros((n_data, 1), dtype=np.int64)

        # if self.use_smart_probing and self.n_probe > 1:
        #     p = -topk_sims.abs().sqrt()
        #     p = torch.softmax(p / self.smart_probing_temperature, dim=-1)
        #
        #     # p_norm = p.norm(dim=-1)
        #     # sqrt_d = self.n_probe ** 0.5
        #     # score = 1 - (p_norm * sqrt_d - 1) / (sqrt_d - 1) - 1e-6
        #     # n_probe_list = torch.ceil(score * (self.n_probe) ).long()
        #
        #     max_n_probe = torch.tensor(self.n_probe, device=self.device)
        #     normalized_entropy = - torch.sum(p * torch.log2(p) / torch.log2(max_n_probe), dim=-1)
        #     n_probe_list = torch.ceil(normalized_entropy * max_n_probe).long()
        # else:
        #     n_probe_list = None
        return cells

    def search_numpy(
        self,
        query_np: 'np.ndarray',
        filter: Dict = {},
        limit: int = 10,
        **kwargs,
    ):
        """Search the index and return distances to the query and ids of the closest documents.

        :param query_np: matrix containing query vectors as rows
        :param filter: the filtering conditions
        :param limit: the number of results to get for each query document in search
        """

        if not self.is_trained:
            raise RuntimeError(f'The indexer is not trained, cannot add new documents')

        dists, doc_ids = self._search_numpy(query_np, filter, limit)
        return dists, doc_ids

    def _search_numpy(self, query_np: 'np.ndarray', filter: Dict = {}, limit: int = 10):
        """Search approximate nearest vectors in different cells, returns distances and ids

        :param query_np: matrix containing query vectors as rows
        :param filter: the filtering conditions
        :param limit: the number of results to get for each query document in search
        """
        cells = self._cell_selection(query_np, limit)
        where_clause, where_params = Filter(filter).parse_where_clause()

        dists, ids = self._search_cells(
            query=query_np,
            cells=cells,
            where_clause=where_clause,
            where_params=where_params,
            limit=limit,
        )
        return dists, ids

    def delete(
        self,
        docs: Union['DocumentArray', List[str]],
        raise_errors_on_not_found: bool = False,
    ):
        """Delete entries from the index by id

        :param raise_errors_on_not_found: whether to raise exception when id not found.
        :param docs: the documents to delete
        """

        super().delete(
            docs if isinstance(docs, list) else docs[:, 'id'], raise_errors_on_not_found
        )

    def clear(self):
        """Clear the whole database"""
        for cell_id in range(self.n_cells):
            self.vec_index(cell_id).reset()
            self.cell_table(cell_id).clear()
            self.doc_store(cell_id).clear()
        self.meta_table.clear()

    def close(self):
        for cell_id in range(self.n_cells):
            self.doc_store(cell_id).close()

    def encode(self, x: 'np.ndarray'):
        n_data, _ = self._sanity_check(x)

        if self._projector_codec:
            x = self._projector_codec.encode(x)

        if self._vq_codec:
            x = self._pq_codec.encode(x)

        return x

    def decode(self, x: 'np.ndarray'):
        assert len(x.shape) == 2
        assert x.shape[1] == self.n_subvectors

        if self._pq_codec:
            x = self._pq_codec.decode(x)

        if self._projector_codec:
            x = self._projector_codec.decode(x)

        return x

    @property
    def params_hash(self):
        model_metas = (
            f'n_dim: {self.n_dim} '
            f'metric: {self.metric} '
            f'n_cells: {self.n_cells} '
            f'n_components: {self.n_components} '
            f'n_subvectors: {self.n_subvectors}'
        )
        return hashlib.md5(f'{model_metas}'.encode()).hexdigest()

    @property
    def model_path(self):
        return self.data_path / f'parameters-{self.params_hash}'

    @property
    def _vq_codec_path(self):
        return self.model_path / f'vq_codec.params'

    @property
    def _pq_codec_path(self):
        return self.model_path / f'pq_codec.params'

    @property
    def _projector_codec_path(self):
        return self.model_path / f'projector_codec.params'

    @property
    def index_hash(self):
        latest_commit = self.meta_table.get_latest_commit()
        date_time = latest_commit[-1] if latest_commit else None
        if date_time:
            if platform.system() == 'Windows':
                return date_time.isoformat('#', 'hours')
            return date_time.isoformat('#', 'seconds')

        return None

    @property
    def index_path(self):
        if self.index_hash:
            return (
                self.data_path
                / f'snapshot-{self.params_hash}'
                / f'{self.index_hash}-SNAPSHOT'
            )
        return None

    @property
    def snapshot_path(self):
        paths = list(
            (self.data_path / f'snapshot-{self.params_hash}').glob(f'*-SNAPSHOT')
        )

        if paths:
            paths = sorted(paths, key=lambda x: x.name)
            return paths[-1]
        return None

    @property
    def remote_store_client(self):
        try:
            import hubble

            os.environ['JINA_AUTH_TOKEN'] = self.token
            client = hubble.Client(max_retries=None, jsonify=True)
            client.get_user_info()
            return client
        except Exception as ex:
            logger.error(f'Not login to hubble yet.')
            raise ex

    def backup(self, target_name: Optional[str] = None, token: Optional[str] = None):
        # file lock will be released when backup to remote, this will
        # release the file lock. And it's only needed in Windows
        # since we need to release file lock before we can access rocksdb files.
        if not target_name:
            logger.info('dump to local ...')
            self.dump()
        else:
            if token is None:
                logger.error(f'back up to remote needs token')
            logger.info(f'dump to remote: {target_name}')
            self.close()
            self._backup_index_to_remote(target_name, token)

    def restore(self, source_name: Optional[str] = None, token: Optional[str] = None):
        # file lock will be released when restore from remote
        if not source_name:
            if self.total_docs > 0:
                logger.info(f'restore Annlite from local')
                self._rebuild_index_from_local()
        else:
            if token is None:
                logger.error(f'restore from remote needs token')
            logger.info(f'restore Annlite from artifact: {source_name}')
            self.close()
            self._rebuild_index_from_remote(source_name, token)

    def dump_model(self):
        logger.info(f'Save the parameters to {self.model_path}')
        self.model_path.mkdir(parents=True, exist_ok=True)
        if self._projector_codec:
            self._projector_codec.dump(self._projector_codec_path)
        if self._vq_codec:
            self._vq_codec.dump(self._vq_codec_path)
        if self._pq_codec:
            self._pq_codec.dump(self._pq_codec_path)

    def dump_index(self):
        import shutil

        logger.info(f'Save the indexer to {self.index_path}')
        try:
            if Path.exists(self.index_path):
                logger.info(
                    f'Index path {self.index_path} already exists, will be '
                    f'overwritten'
                )
                shutil.rmtree(self.index_path)
            self.index_path.mkdir(parents=True)

            for cell_id in range(self.n_cells):
                self.vec_index(cell_id).dump(self.index_path / f'cell_{cell_id}.hnsw')
                self.cell_table(cell_id).dump(self.index_path / f'cell_{cell_id}.db')
        except Exception as ex:
            logger.error(f'Failed to dump the indexer, {ex!r}')

            if self.index_path:
                shutil.rmtree(self.index_path)

    def dump(self):
        self.dump_model()
        self.dump_index()

    def _backup_index_to_remote(self, target_name: str, token: str):

        self.dump()

        from .hubble_tools import Uploader

        self.token = token
        client = self.remote_store_client
        uploader = Uploader(size_limit=self.size_limit, client=client)

        for cell_id in range(self.n_cells):
            # upload database
            uploader.upload_directory(
                Path(self.data_path) / f'cell_{cell_id}',
                target_name=target_name,
                type='database',
                cell_id=cell_id,
            )

            # upload hnsw file
            uploader.upload_file(
                Path(self.index_path) / f'cell_{cell_id}.hnsw',
                target_name=target_name,
                type='hnsw',
                cell_id=cell_id,
            )

            # upload cell_table
            uploader.upload_file(
                Path(self.index_path) / f'cell_{cell_id}.db',
                target_name=target_name,
                type='cell_table',
                cell_id=cell_id,
            )

        # upload meta table
        uploader.upload_file(
            Path(self.data_path) / 'metas.db',
            target_name=target_name,
            type='meta_table',
            cell_id='all',
        )

        # upload training model
        uploader.archive_and_upload(
            target_name,
            'model',
            'model.zip',
            'all',
            self.model_path.parent,
            str(self.model_path.name),
        )

    def _rebuild_index_from_local(self):
        if self.snapshot_path:
            logger.info(f'Load the indexer from snapshot {self.snapshot_path}')
            for cell_id in range(self.n_cells):
                self.vec_index(cell_id).load(
                    self.snapshot_path / f'cell_{cell_id}.hnsw'
                )
                self.cell_table(cell_id).load(self.snapshot_path / f'cell_{cell_id}.db')
        else:
            logger.info(f'Rebuild the indexer from scratch')
            for cell_id in range(self.n_cells):
                cell_size = self.doc_store(cell_id).size

                if cell_size == 0:
                    continue  # skip empty cell

                logger.debug(
                    f'Rebuild the index of cell-{cell_id} ({cell_size} docs)...'
                )
                for docs in self.documents_generator(cell_id, batch_size=10240):
                    x = to_numpy_array(docs.embeddings)

                    assigned_cells = np.ones(len(docs), dtype=np.int64) * cell_id
                    super().insert(x, assigned_cells, docs, only_index=True)
                logger.debug(f'Rebuild the index of cell-{cell_id} done')
        if self.model_path:
            logger.info(f'Load the model from {self.model_path}')
            self._reload_models()

    def _rebuild_index_from_remote(self, source_name: str, token: str):
        import shutil

        from .hubble_tools import Merger

        self.token = token
        client = self.remote_store_client
        art_list = client.list_artifacts(
            filter={'metaData.name': source_name}, pageSize=100
        )
        if len(art_list['data']) == 0:
            logger.info(f'The indexer `{source_name}` not found. ')
        else:
            logger.info(f'Load the indexer `{source_name}` from remote store')

            restore_path = self.data_path / 'restore'
            merger = Merger(restore_path=restore_path, client=client)

            for cell_id in range(self.n_cells):
                # download hnsw files and merge and load
                logger.info(f'Load the hnsw `{source_name}` from remote store')

                hnsw_ids = merger.get_artifact_ids(
                    art_list, type='hnsw', cell_id=cell_id
                )
                merger.download(ids=hnsw_ids, download_folder=f'hnsw_{cell_id}')
                if len(hnsw_ids) > 1:
                    merger.merge_file(
                        inputdir=restore_path / f'hnsw_{cell_id}',
                        outputdir=restore_path / f'hnsw_{cell_id}',
                        outputfilename=Path(f'cell_{cell_id}.hnsw'),
                    )
                self.vec_index(cell_id).load(
                    restore_path / f'hnsw_{cell_id}' / f'cell_{cell_id}.hnsw'
                )
                shutil.rmtree(restore_path / f'hnsw_{cell_id}')

                # download cell_table files and merge and load
                logger.info(f'Load the cell_table `{source_name}` from remote store')

                cell_table_ids = merger.get_artifact_ids(
                    art_list, type='cell_table', cell_id=cell_id
                )
                merger.download(
                    ids=cell_table_ids, download_folder=f'cell_table_{cell_id}'
                )
                if len(cell_table_ids) > 1:
                    merger.merge_file(
                        inputdir=restore_path / f'cell_table_{cell_id}',
                        outputdir=restore_path / f'cell_table_{cell_id}',
                        outputfilename=Path(f'cell_{cell_id}.db'),
                    )

                self.cell_table(cell_id).load(
                    restore_path / f'cell_table_{cell_id}' / f'cell_{cell_id}.db'
                )
                shutil.rmtree(restore_path / f'cell_table_{cell_id}')

                # download database files and rebuild
                logger.info(f'Load the database `{source_name}` from remote store')

                database_ids = merger.get_artifact_ids(art_list, type='database')
                merger.download(ids=database_ids, download_folder='database')
                for zip_file in list((restore_path / 'database').iterdir()):
                    # default has only one cell
                    shutil.unpack_archive(zip_file, self.data_path / f'cell_{cell_id}')
                    for f in list(
                        (
                            self.data_path
                            / f'cell_{cell_id}'
                            / zip_file.name.split('.zip')[0]
                        ).iterdir()
                    ):
                        origin_database_path = (
                            self.data_path / f'cell_{cell_id}' / f.name
                        )
                        if origin_database_path.exists():
                            origin_database_path.unlink()
                        f.rename(self.data_path / f'cell_{cell_id}' / f.name)
                    shutil.rmtree(
                        self.data_path
                        / f'cell_{cell_id}'
                        / zip_file.name.split('.zip')[0]
                    )
                    Path(zip_file).unlink()
                self._rebuild_database()

            # download meta_table files
            logger.info(f'Load the meta_table `{source_name}` from remote store')

            meta_table_ids = merger.get_artifact_ids(art_list, type='meta_table')
            merger.download(ids=meta_table_ids, download_folder='meta_table')

            if len(meta_table_ids) > 1:
                merger.merge_file(
                    inputdir=restore_path / 'meta_table',
                    outputdir=self.data_path,
                    outputfilename=Path('metas.db'),
                )
            else:
                mata_table_file = restore_path / 'meta_table' / 'metas.db'
                if platform.system() == 'Windows':
                    origin_metas_path = self.data_path / 'metas.db'
                    if origin_metas_path.exists():
                        self._meta_table.close()
                        origin_metas_path.unlink()
                mata_table_file.rename(self.data_path / 'metas.db')
                if platform.system() == 'Windows':
                    from .storage.table import MetaTable

                    self._meta_table = MetaTable(
                        'metas', data_path=self.data_path, in_memory=False
                    )
            shutil.rmtree(restore_path / 'meta_table')

            # download model files
            logger.info(f'Load the model `{source_name}` from remote store')
            file_name = str(self.model_path.parent / f'{source_name}_model.zip')
            model_id = [
                art['_id']
                for art in art_list['data']
                if 'model' in art['metaData']['type']
            ]
            assert len(model_id) == 1
            client.download_artifact(
                id=model_id[0],
                f=file_name,
                show_progress=True,
            )
            shutil.unpack_archive(file_name, self.model_path.parent)
            self._reload_models()
            Path(file_name).unlink()

            shutil.rmtree(restore_path)

    @property
    def is_trained(self):
        if self._projector_codec and (not self._projector_codec.is_trained):
            return False
        if self._vq_codec and (not self._vq_codec.is_trained):
            return False
        if self._pq_codec and (not self._pq_codec.is_trained):
            return False
        return True

    def _reload_models(self):
        if self._projector_codec_path.exists():
            self._projector_codec = ProjectorCodec.load(self._projector_codec_path)
        if self._vq_codec_path.exists():
            self._vq_codec = VQCodec.load(self._vq_codec_path)
        if self._pq_codec_path.exists():
            self._pq_codec = PQCodec.load(self._pq_codec_path)

    @property
    def use_smart_probing(self):
        return self._use_smart_probing

    @use_smart_probing.setter
    def use_smart_probing(self, value):
        assert type(value) is bool
        self._use_smart_probing = value

    @property
    def stat(self):
        """Get information on status of the indexer."""
        return {
            'total_docs': self.total_docs,
            'index_size': self.index_size,
            'n_cells': self.n_cells,
            'n_dim': self.n_dim,
            'n_components': self.n_components,
            'metric': self.metric.name,
            'is_trained': self.is_trained,
        }

    # @property
    # def smart_probing_temperature(self):
    #     return self._smart_probing_temperature
    #
    # @smart_probing_temperature.setter
    # def smart_probing_temperature(self, value):
    #     assert value > 0
    #     assert self.use_smart_probing, 'set use_smart_probing to True first'
    #     self._smart_probing_temperature = value



from typing import Tuple

import numpy as np


def l2_normalize(x: 'np.ndarray', eps: float = np.finfo(np.float32).eps):
    """Scale input vectors individually to unit norm.

    :param x: The data to normalize
    :param eps: a small jitter to avoid divde by zero
    :return: Normalized input X
    """

    norms = np.einsum('ij,ij->i', x, x)
    np.sqrt(norms, norms)
    constant_mask = norms < 10 * eps
    norms[constant_mask] = 1.0
    return x / norms[:, np.newaxis]


def cosine(
    x_mat: 'np.ndarray', y_mat: 'np.ndarray', eps: float = np.finfo(np.float32).eps
) -> 'np.ndarray':
    """Cosine distance between each row in x_mat and each row in y_mat.

    :param x_mat: np.ndarray with ndim=2
    :param y_mat: np.ndarray with ndim=2
    :param eps: a small jitter to avoid divde by zero
    :return: np.ndarray  with ndim=2
    """
    return 1 - np.clip(
        (np.dot(x_mat, y_mat.T) + eps)
        / (
            np.outer(np.linalg.norm(x_mat, axis=1), np.linalg.norm(y_mat, axis=1)) + eps
        ),
        -1,
        1,
    )


def sqeuclidean(x_mat: 'np.ndarray', y_mat: 'np.ndarray') -> 'np.ndarray':
    """Squared Euclidean distance between each row in x_mat and each row in y_mat.
    :param x_mat: np.ndarray with ndim=2
    :param y_mat: np.ndarray with ndim=2
    :return: np.ndarray with ndim=2
    """
    return (
        np.sum(y_mat**2, axis=1)
        + np.sum(x_mat**2, axis=1)[:, np.newaxis]
        - 2 * np.dot(x_mat, y_mat.T)
    )


def euclidean(x_mat: 'np.ndarray', y_mat: 'np.ndarray') -> 'np.ndarray':
    """Euclidean distance between each row in x_mat and each row in y_mat.

    :param x_mat:  scipy.sparse like array with ndim=2
    :param y_mat:  scipy.sparse like array with ndim=2
    :return: np.ndarray  with ndim=2
    """
    return np.sqrt(sqeuclidean(x_mat, y_mat))


def pdist(
    x_mat: 'np.ndarray',
    metric: str,
) -> 'np.ndarray':
    """Computes Pairwise distances between observations in n-dimensional space.

    :param x_mat: Union['np.ndarray','scipy.sparse.csr_matrix', 'scipy.sparse.coo_matrix'] of ndim 2
    :param metric: string describing the metric type
    :return: np.ndarray of ndim 2
    """
    return cdist(x_mat, x_mat, metric)


def cdist(x_mat: 'np.ndarray', y_mat: 'np.ndarray', metric: str) -> 'np.ndarray':
    """Computes the pairwise distance between each row of X and each row on Y according to `metric`.
    - Let `n_x = x_mat.shape[0]`
    - Let `n_y = y_mat.shape[0]`
    - Returns a matrix `dist` of shape `(n_x, n_y)` with `dist[i,j] = metric(x_mat[i], y_mat[j])`.
    :param x_mat: numpy or scipy array of ndim 2
    :param y_mat: numpy or scipy array of ndim 2
    :param metric: string describing the metric type
    :return: np.ndarray of ndim 2
    """
    dists = {'cosine': cosine, 'sqeuclidean': sqeuclidean, 'euclidean': euclidean}[
        metric
    ](x_mat, y_mat)

    return dists


def top_k(
    values: 'np.ndarray', k: int, descending: bool = False
) -> Tuple['np.ndarray', 'np.ndarray']:
    """Finds values and indices of the k largest entries for the last dimension.

    :param values: array of distances
    :param k: number of values to retrieve
    :param descending: find top k biggest values
    :return: indices and distances
    """
    if descending:
        values = -values

    if k >= values.shape[1]:
        idx = values.argsort(axis=1)[:, :k]
        values = np.take_along_axis(values, idx, axis=1)
    else:
        idx_ps = values.argpartition(kth=k, axis=1)[:, :k]
        values = np.take_along_axis(values, idx_ps, axis=1)
        idx_fs = values.argsort(axis=1)
        idx = np.take_along_axis(idx_ps, idx_fs, axis=1)
        values = np.take_along_axis(values, idx_fs, axis=1)

    if descending:
        values = -values

    return values, idx



from typing import List, Optional

import numpy as np

from ...math import top_k
from ..codec.pq import PQCodec
from .flat_index import FlatIndex


# TODO: deprecated this index
class PQIndex(FlatIndex):  # pragma: no cover
    def __init__(
        self,
        dim: int,
        pq_codec: PQCodec,
        **kwargs,
    ):
        assert pq_codec is not None
        self._dense_dim = dim
        super(PQIndex, self).__init__(
            pq_codec.n_subvectors, dtype=pq_codec.code_dtype, **kwargs
        )
        self._pq_codec = pq_codec

    def add_with_ids(self, x: np.ndarray, ids: List[int]):
        x = self._pq_codec.encode(x)
        super(PQIndex, self).add_with_ids(x, ids)

    def search(
        self, x: np.ndarray, limit: int = 10, indices: Optional[np.ndarray] = None
    ):
        _dim = x.shape[-1]
        assert (
            _dim == self._pq_codec.dim
        ), f'the query embedding dimension does not match with index dimension: {_dim} vs {self.dim}'

        precomputed = self._pq_codec.precompute_adc(x)

        codes = self._data
        data_idx = np.arange(self._capacity)

        if indices is not None:
            codes = self._data[indices]
            data_idx = data_idx[indices]

        dists = precomputed.adist(codes)  # (10000, )
        dists = np.expand_dims(dists, axis=0)

        dists, ids = top_k(dists, limit, descending=False)

        # TODO: change the shape of return
        ids = ids[0]
        if indices is not None:
            ids = data_idx[ids]

        return dists[0], ids



import time
import warnings
from pathlib import Path
from typing import Dict, List, Union

from docarray import Document, DocumentArray
from rocksdict import Options, Rdict, ReadOptions, WriteBatch, WriteOptions


class DocStorage:
    """The backend storage engine of Documents"""

    def __init__(
        self,
        path: Union[str, Path],
        serialize_config: Dict = {},
        create_if_missing: bool = True,
        **kwargs,
    ):
        self._path = str(path)
        self._serialize_config = serialize_config

        self._kwargs = kwargs

        self._init_db(create_if_missing=create_if_missing, **self._kwargs)

    def _init_db(self, create_if_missing: bool = True, **kwargs):
        opt = Options(raw_mode=True)

        opt.optimize_for_point_lookup(1024)

        opt.set_inplace_update_support(True)
        opt.set_allow_concurrent_memtable_write(False)

        # configure mem-table to a large value (256 MB)
        opt.set_write_buffer_size(0x10000000)

        # 256 MB file size
        opt.set_target_file_size_base(0x10000000)

        # # set to plain-table for better performance
        # opt.set_plain_table_factory(PlainTableFactoryOptions())

        opt.create_if_missing(create_if_missing)

        self._db = Rdict(path=self._path, options=opt)

        # get the size of the database, if it is not created, set it to 0
        self._size = len(list(self._db.keys()))

        self._is_closed = False

    def insert(self, docs: 'DocumentArray'):
        write_batch = WriteBatch(raw_mode=True)
        write_opt = WriteOptions()
        write_opt.set_sync(True)
        batch_size = 0
        for doc in docs:
            write_batch.put(doc.id.encode(), doc.to_bytes(**self._serialize_config))
            batch_size += 1
        self._db.write(write_batch, write_opt=write_opt)
        self._size += batch_size

    def update(self, docs: 'DocumentArray'):
        write_batch = WriteBatch(raw_mode=True)
        write_opt = WriteOptions()
        write_opt.set_sync(True)
        for doc in docs:
            key = doc.id.encode()
            if key not in self._db:
                raise ValueError(f'The Doc ({doc.id}) does not exist in database!')

            write_batch.put(key, doc.to_bytes(**self._serialize_config))
        self._db.write(write_batch, write_opt=write_opt)

    def delete(self, doc_ids: List[str]):
        write_batch = WriteBatch(raw_mode=True)
        write_opt = WriteOptions()
        write_opt.set_sync(True)
        for doc_id in doc_ids:
            write_batch.delete(doc_id.encode())
        self._db.write(write_batch, write_opt=write_opt)
        self._size -= len(doc_ids)

    def get(self, doc_ids: Union[str, list]) -> DocumentArray:
        docs = DocumentArray()
        if isinstance(doc_ids, str):
            doc_ids = [doc_ids]

        for doc_bytes in self._db[[k.encode() for k in doc_ids]]:
            if doc_bytes:
                docs.append(Document.from_bytes(doc_bytes, **self._serialize_config))

        return docs

    def clear(self):
        if self._is_closed:
            warnings.warn(
                '`DocStorage` had been closed already, will skip this close operation.'
            )
        else:
            self._db.close()
        self._db.destroy(self._path)

        # re-initialize the database for the next usage
        self._init_db(create_if_missing=True, **self._kwargs)

    def close(self):
        if self._is_closed:
            warnings.warn(
                '`DocStorage` had been closed already, will skip this close operation.'
            )
            return
        self._db.flush(wait=True)
        self._db.close()
        self._is_closed = True

    def __len__(self):
        return self._size

    @property
    def stat(self):
        return {'entries': len(self)}

    @property
    def size(self):
        return self.stat['entries']

    @property
    def last_transaction_id(self):
        return self._db.latest_sequence_number()

    def batched_iterator(self, batch_size: int = 1, **kwargs) -> 'DocumentArray':
        count = 0
        docs = DocumentArray()

        read_opt = ReadOptions(raw_mode=True)

        for value in self._db.values(read_opt=read_opt):
            doc = Document.from_bytes(value, **self._serialize_config)
            docs.append(doc)
            count += 1

            if count == batch_size:
                yield docs
                count = 0
                docs = DocumentArray()

        if count > 0:
            yield docs



import os
import platform
import sys
from distutils.sysconfig import get_python_inc

import numpy as np
import pybind11
import setuptools
from Cython.Build import cythonize
from setuptools import Extension, find_packages, setup
from setuptools.command.build_ext import build_ext

if sys.version_info < (3, 7, 0):
    raise OSError(f'annlite requires Python 3.7+, but yours is {sys.version}')

include_dirs = [pybind11.get_include(), np.get_include()]

libraries = []
extra_objects = []

try:
    pkg_name = 'annlite'
    libinfo_py = os.path.join(pkg_name, '__init__.py')
    libinfo_content = open(libinfo_py, 'r', encoding='utf8').readlines()
    version_line = [l.strip() for l in libinfo_content if l.startswith('__version__')][
        0
    ]
    exec(version_line)  # produce __version__
except FileNotFoundError:
    __version__ = '0.0.0'

try:
    with open('README.md', encoding='utf8') as fp:
        _long_description = fp.read()
except FileNotFoundError:
    _long_description = ''

try:
    with open('requirements.txt') as f:
        base_deps = f.read().splitlines()

    # remove blank lines and comments
    base_deps = [
        x.strip()
        for x in base_deps
        if ((x.strip()[0] != '#') and (len(x.strip()) > 3) and '-e git://' not in x)
    ]
except FileNotFoundError:
    base_deps = []

COMPILER_DIRECTIVES = {
    'language_level': -3,
    'embedsignature': True,
    'annotation_typing': False,
}

ext_modules = [
    Extension(
        'annlite.hnsw_bind',
        ['./bindings/hnsw_bindings.cpp'],
        include_dirs=include_dirs + ['./include/hnswlib'],
        libraries=libraries,
        language='c++',
        extra_objects=extra_objects,
    )
] + cythonize(
    [
        Extension(
            'annlite.pq_bind',
            ['./bindings/pq_bindings.pyx'],
            include_dirs=include_dirs + [get_python_inc(plat_specific=True)],
            libraries=libraries,
            language='c++',
            extra_objects=extra_objects,
        ),
    ],
    compiler_directives=COMPILER_DIRECTIVES,
)
#
# ext_modules = cythonize([
#                 Extension(
#                     'pqlite.pq_bind',
#                     ['./bindings/pq_bindings.pyx'],
#                     include_dirs=include_dirs + [get_python_inc(plat_specific=True)],
#                     libraries=libraries,
#                     language='c++',
#                     extra_objects=extra_objects,
#                 ),
#             ], compiler_directives=COMPILER_DIRECTIVES)

# As of Python 3.6, CCompiler has a `has_flag` method.
# cf http://bugs.python.org/issue26689
def has_flag(compiler, flagname):
    """Return a boolean indicating whether a flag name is supported on
    the specified compiler.
    """
    import tempfile

    with tempfile.NamedTemporaryFile('w', suffix='.cpp') as f:
        f.write('int main (int argc, char **argv) { return 0; }')
        try:
            compiler.compile([f.name], extra_postargs=[flagname])
        except setuptools.distutils.errors.CompileError:
            return False
    return True


def cpp_flag(compiler):
    """Return the -std=c++[11/14] compiler flag.
    The c++14 is prefered over c++11 (when it is available).
    """
    if has_flag(compiler, '-std=c++14'):
        return '-std=c++14'
    elif has_flag(compiler, '-std=c++11'):
        return '-std=c++11'
    else:
        raise RuntimeError(
            'Unsupported compiler -- at least C++11 support ' 'is needed!'
        )


class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""

    c_opts = {
        'msvc': ['/EHsc', '/openmp', '/O2'],
        'unix': ['-O3', '-march=native'],  # , '-w'
    }
    link_opts = {
        'unix': [],
        'msvc': [],
    }

    if sys.platform == 'darwin':
        if platform.processor() in ('arm64', 'arm'):
            c_opts['unix'].remove('-march=native')
            # thanks to @https://github.com/drkeoni
            # https://github.com/nmslib/nmslib/issues/476#issuecomment-876094529
            c_opts['unix'].append('-mcpu=apple-a14')
        c_opts['unix'] += ['-stdlib=libc++', '-mmacosx-version-min=10.7']
        link_opts['unix'] += ['-stdlib=libc++', '-mmacosx-version-min=10.7']
    else:
        c_opts['unix'].append('-fopenmp')
        link_opts['unix'].extend(['-fopenmp', '-pthread'])

    def build_extensions(self):
        ct = self.compiler.compiler_type
        opts = self.c_opts.get(ct, [])
        if ct == 'unix':
            opts.append('-DVERSION_INFO="%s"' % self.distribution.get_version())
            opts.append(cpp_flag(self.compiler))
            # if has_flag(self.compiler, '-fvisibility=hidden'):
            #     opts.append('-fvisibility=hidden')
        elif ct == 'msvc':
            opts.append('/DVERSION_INFO=\\"%s\\"' % self.distribution.get_version())

        for ext in self.extensions:
            ext.extra_compile_args.extend(opts)
            ext.extra_link_args.extend(self.link_opts.get(ct, []))

        build_ext.build_extensions(self)


extras = {}
extras['test'] = [
    'pytest',
    'black',
    'pytest-timeout',
    'pytest-mock',
    'pytest-cov',
    'pytest-repeat',
    'pytest-reraise',
    'pytest-custom_exit_code',
    'jina',
]

# for e in ext_modules:
#     e.cython_directives = COMPILER_DIRECTIVES

setup(
    name='annlite',
    version=__version__,
    description='Fast and Light Approximate Nearest Neighbor Search Library integrated with the Jina Ecosystem',
    long_description=_long_description,
    long_description_content_type='text/markdown',
    author='Jina AI',
    author_email='team@jina.ai',
    url='https://github.com/jina-ai/annlite',
    download_url='https://github.com/jina-ai/pqlite/tags',
    license='Apache License 2.0',
    extras_require=extras,
    ext_modules=ext_modules,
    cmdclass={'build_ext': BuildExt},
    package_data={'bindings': ['*.pyx', '*.pxd', '*.pxi']},
    install_requires=base_deps,
    setup_requires=['setuptools>=18.0', 'wheel', 'cython', 'build'],
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    python_requires='>=3.7',
    packages=find_packages(
        exclude=[
            '*.tests',
            '*.tests.*',
            'tests.*',
            'tests',
            'test',
            'docs',
            'src',
            'executor',
        ]
    ),
    zip_safe=False,
    keywords='product-quantization approximate-nearest-neighbor hnsw',
)



import datetime
import sqlite3
import threading
from pathlib import Path
from typing import TYPE_CHECKING, Any, List, Optional, Tuple, Union

import numpy as np

if TYPE_CHECKING:
    from docarray import DocumentArray

sqlite3.register_adapter(np.int64, lambda x: int(x))
sqlite3.register_adapter(np.int32, lambda x: int(x))

COLUMN_TYPE_MAPPING = {
    float: 'FLOAT',
    int: 'INTEGER',
    bool: 'INTEGER',
    str: 'TEXT',
    bytes.__class__: 'BLOB',
    bytes: 'BLOB',
    memoryview: 'BLOB',
    # datetime.datetime: 'TEXT',
    # datetime.date: 'TEXT',
    # datetime.time: 'TEXT',
    None.__class__: 'TEXT',
    # SQLite explicit types
    'TEXT': 'TEXT',
    'INTEGER': 'INTEGER',
    'FLOAT': 'FLOAT',
    'BLOB': 'BLOB',
    'text': 'TEXT',
    'integer': 'INTEGER',
    'float': 'FLOAT',
    'blob': 'BLOB',
}

# If numpy is available, add more types
if np:
    COLUMN_TYPE_MAPPING.update(
        {
            np.int8: 'INTEGER',
            np.int16: 'INTEGER',
            np.int32: 'INTEGER',
            np.int64: 'INTEGER',
            np.uint8: 'INTEGER',
            np.uint16: 'INTEGER',
            np.uint32: 'INTEGER',
            np.uint64: 'INTEGER',
            np.float16: 'FLOAT',
            np.float32: 'FLOAT',
            np.float64: 'FLOAT',
        }
    )


def _converting(value: Any) -> str:
    if isinstance(value, bool):
        if value:
            return 1
        else:
            return 0

    return str(value)


def time_now():
    return datetime.datetime.utcnow()


def _get_table_names(
    conn: 'sqlite3.Connection', fts4: bool = False, fts5: bool = False
) -> List[str]:
    """A list of string table names in this database."""
    where = ["type = 'table'"]
    if fts4:
        where.append("sql like '%USING FTS4%'")
    if fts5:
        where.append("sql like '%USING FTS5%'")
    sql = 'select name from sqlite_master where {}'.format(' AND '.join(where))
    return [r[0] for r in conn.execute(sql).fetchall()]


class Table:
    def __init__(
        self,
        name: str,
        data_path: Optional[Union[Path, str]] = None,
        detect_types: int = 0,
        in_memory: bool = True,
    ):
        if in_memory:
            self._conn_name = ':memory:'
        else:
            if isinstance(data_path, str):
                data_path = Path(data_path)
            self._conn_name = data_path / f'{name}.db'
        self._name = name

        self.detect_types = detect_types

        self._conn = sqlite3.connect(
            self._conn_name, detect_types=detect_types, check_same_thread=False
        )
        self._conn_lock = threading.Lock()

    def execute(self, sql: str, commit: bool = True):
        self._conn.execute(sql)
        if commit:
            self.commit()

    def execute_many(self, sql: str, parameters: List[Tuple], commit: bool = True):
        self._conn.executemany(sql, parameters)
        if commit:
            self.commit()

    def commit(self):
        self._conn.commit()

    def create_table(self):
        ...

    def drop_table(self):
        self._conn.execute(f'DROP table {self.name}')
        self._conn.commit()

    def clear(self):
        """Drop the table and create a new one"""
        self.drop_table()
        self.create_table()

    def load(self, data_file: Union[str, Path]):
        disk_db = sqlite3.connect(data_file, detect_types=self.detect_types)
        disk_db.backup(self._conn)
        disk_db.close()

    def dump(self, data_file: Union[str, Path]):
        backup_db = sqlite3.connect(data_file, detect_types=self.detect_types)
        self._conn.backup(backup_db)
        backup_db.close()

    def close(self):
        self._conn.close()

    @property
    def name(self):
        return self._name

    @property
    def schema(self):
        """SQL schema for this database"""
        result = []
        for row in self._conn.execute(
            f'''PRAGMA table_info("{self.name}")'''
        ).fetchall():
            result.append(', '.join([str(_) for _ in row]))
        return '\n'.join(result)


class CellTable(Table):
    def __init__(
        self,
        name: str,
        columns: Optional[List[tuple]] = None,
        in_memory: bool = True,
        data_path: Optional[Path] = None,
        lazy_create: bool = False,
    ):
        super().__init__(name, data_path=data_path, in_memory=in_memory)

        self._columns = []
        self._indexed_keys = set()

        if columns is not None:
            for name, dtype in columns:
                self.add_column(name, dtype, True)
        if not lazy_create:
            self.create_table()

    @property
    def columns(self) -> List[str]:
        return ['_id', '_doc_id'] + [c.split()[0] for c in self._columns]

    def existed(self):
        return self.name in _get_table_names(self._conn)

    def add_column(self, name: str, dtype: str, create_index: bool = True):
        self._columns.append(f'{name} {COLUMN_TYPE_MAPPING[dtype]}')
        if create_index:
            self._indexed_keys.add(name)

    def create_index(self, column: str, commit: bool = True):
        sql_statement = f'''CREATE INDEX idx_{column}_
                            ON {self.name}(_deleted, {column})'''
        self._conn.execute(sql_statement)

        if commit:
            self._conn.commit()

    def create_table(self):
        sql = f'''CREATE TABLE {self.name}
                    (_id INTEGER PRIMARY KEY AUTOINCREMENT,
                     _doc_id TEXT NOT NULL UNIQUE,
                     _deleted NUMERIC DEFAULT 0'''
        if len(self._columns) > 0:
            sql += ', ' + ', '.join(self._columns)
        sql += ')'
        self._conn.execute(sql)

        sql_statement = f'''CREATE INDEX idx__delete_
                                ON {self.name}(_deleted)'''
        self._conn.execute(sql_statement)

        for name in self._indexed_keys:
            self.create_index(name, commit=False)
        self._conn.commit()

    def insert(
        self,
        docs: 'DocumentArray',
        commit: bool = True,
    ) -> List[int]:
        """Add a single record into the table.

        :param docs: The list of dict docs
        :param commit: If set, commit is applied
        """
        sql_template = 'INSERT INTO {table}({columns}) VALUES ({placeholders});'

        column_names = self.columns[1:]
        columns = ', '.join(column_names)
        placeholders = ', '.join('?' for c in column_names)
        sql = sql_template.format(
            table=self.name, columns=columns, placeholders=placeholders
        )

        values = []
        docs_size = 0
        for doc in docs:
            doc_value = tuple(
                [doc.id]
                + [
                    _converting(doc.tags[c]) if c in doc.tags else None
                    for c in self.columns[2:]
                ]
            )
            values.append(doc_value)
            docs_size += 1

        with self._conn_lock:
            cursor = self._conn.cursor()
            if docs_size > 1:
                cursor.executemany(sql, values[:-1])

            cursor.execute(sql, values[-1])
            last_row_id = cursor.lastrowid
            row_ids = list(range(last_row_id - len(docs), last_row_id))

            if commit:
                self._conn.commit()

        return row_ids

    def query(
        self,
        where_clause: str = '',
        where_params: Tuple = (),
        limit: int = -1,
        offset: int = 0,
        order_by: Optional[str] = None,
        ascending: bool = True,
    ) -> List[int]:
        """Query the records which matches the given conditions

        :param where_clause: where clause for query
        :param where_params: where parameters for query
        :param limit: limit the number of results
        :param offset: offset the number of results
        :param order_by: order the results by the given column
        :param ascending: order the results in ascending or descending order
        :return: offsets list of matched docs
        """

        where_conds = ['_deleted = ?']
        if where_clause:
            where_conds.append(where_clause)
        where = ' and '.join(where_conds)

        _order_by = f'{order_by or "_id"} {"ASC" if ascending else "DESC"}'
        _limit = f'LIMIT {limit}' if limit > 0 else ''
        _offset = f'OFFSET {offset}' if offset > 0 else ''

        sql = f'SELECT _id from {self.name} WHERE {where} ORDER BY {_order_by} {_limit} {_offset}'

        params = (0,) + tuple([_converting(p) for p in where_params])

        # # EXPLAIN SQL query
        # for row in self._conn.execute('EXPLAIN QUERY PLAN ' + sql, params):
        #     print(row)

        # Use `row_factor`
        # https://docs.python.org/3.6/library/sqlite3.html#sqlite3.Connection.row_factory
        def _offset_factory(_, record):
            return record[0] - 1

        self._conn.row_factory = _offset_factory

        cursor = self._conn.cursor()
        offsets = cursor.execute(sql, params).fetchall()
        self._conn.row_factory = None
        return offsets if offsets else []

    def delete(self, doc_ids: List[str]):
        """Delete the docs

        :param doc_ids: The IDs of docs
        """
        sql = f'UPDATE {self.name} SET _deleted = 1 WHERE _doc_id = ?'
        self._conn.executemany(sql, doc_ids)
        self._conn.commit()

    def get_docid_by_offset(self, offset: int):
        sql = f'SELECT _doc_id from {self.name} WHERE _id = ? and _deleted = 0 LIMIT 1;'
        result = self._conn.execute(sql, (offset + 1,)).fetchone()
        if result:
            return result[0]
        return None

    def delete_by_offset(self, offset: int):
        """Delete the doc with specific offset

        :param offset: The offset of the doc
        """
        sql = f'UPDATE {self.name} SET _deleted = 1 WHERE _id = ?'
        self._conn.execute(sql, (offset + 1,))
        self._conn.commit()

    def undo_delete_by_offset(self, offset: int):
        sql = f'UPDATE {self.name} SET _deleted = 0 WHERE _id = ?'
        self._conn.execute(sql, (offset + 1,))
        self._conn.commit()

    def exist(self, doc_id: str):
        sql = f'SELECT count(*) from {self.name} WHERE _deleted = 0 and _doc_id = ?;'
        return self._conn.execute(sql, (doc_id,)).fetchone()[0] > 0

    def count(self, where_clause: str = '', where_params: Tuple = ()):
        """Return the total number of records which match with the given conditions.
        :param where_clause: where clause for query
        :param where_params: where parameters for query
        :return: the total number of matched records
        """

        if where_clause:
            sql = 'SELECT count(_id) from {table} WHERE {where} LIMIT 1;'
            where = f'_deleted = ? and {where_clause}'
            sql = sql.format(table=self.name, where=where)

            params = (0,) + tuple([_converting(p) for p in where_params])

            # # EXPLAIN SQL query
            # for row in self._conn.execute('EXPLAIN QUERY PLAN ' + sql, params):
            #     print(row)

            return self._conn.execute(sql, params).fetchone()[0]
        else:
            sql = f'SELECT MAX(_id) from {self.name} LIMIT 1;'
            result = self._conn.execute(sql).fetchone()
            if result[0]:
                return result[0] - self.deleted_count()
            return 0

    def deleted_count(self):
        """Return the total number of record what is marked as soft-deleted."""
        sql = f'SELECT count(_id) from {self.name} WHERE _deleted = 1 LIMIT 1'
        return self._conn.execute(sql).fetchone()[0]

    @property
    def size(self):
        return self.count()


class MetaTable(Table):
    def __init__(
        self,
        name: str = 'meta',
        data_path: Optional[Path] = None,
        in_memory: bool = False,
    ):
        super().__init__(
            name,
            data_path=data_path,
            detect_types=sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES,
            in_memory=in_memory,
        )
        self.create_table()

    def create_table(self):
        sql = f'''CREATE TABLE if not exists {self.name}
                        (_doc_id TEXT NOT NULL PRIMARY KEY,
                         cell_id INTEGER NOT NULL,
                         offset INTEGER NOT NULL,
                         time_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
                         _deleted NUMERIC DEFAULT 0)'''

        self._conn.execute(sql)

        self._conn.execute(
            f'CREATE INDEX if not exists idx_time_at_ ON {self.name}(time_at)'
        )
        self._conn.execute(
            f'CREATE INDEX if not exists idx__delete_ ON {self.name}(_deleted)'
        )

        self._conn.commit()

    def iter_addresses(
        self, time_since: 'datetime.datetime' = datetime.datetime(2020, 2, 2, 0, 0)
    ):
        sql = f'SELECT _doc_id, cell_id, offset from {self.name} WHERE time_at >= ? AND _deleted = 0 ORDER BY time_at ASC;'

        cursor = self._conn.cursor()
        for doc_id, cell_id, offset in cursor.execute(sql, (time_since,)):
            yield doc_id, cell_id, offset

    def get_latest_commit(self):
        sql = f'SELECT _doc_id, cell_id, offset, time_at from {self.name} ORDER BY time_at DESC LIMIT 1;'

        cursor = self._conn.execute(sql)
        row = cursor.fetchone()
        return row

    def get_address(self, doc_id: str):
        sql = f'SELECT cell_id, offset from {self.name} WHERE _doc_id = ? AND _deleted = 0 LIMIT 1;'
        cursor = self._conn.execute(sql, (doc_id,))
        row = cursor.fetchone()
        return (row[0], row[1]) if row else (None, None)

    def delete_address(self, doc_id: str, commit: bool = True):
        sql = f'UPDATE {self.name} SET _deleted = 1, time_at = ? WHERE _doc_id = ?'
        self._conn.execute(
            sql,
            (
                time_now(),
                doc_id,
            ),
        )
        print(f'Deleted {doc_id} at: {time_now()}')
        if commit:
            self._conn.commit()

    def add_address(self, doc_id: str, cell_id: int, offset: int, commit: bool = True):
        sql = f'INSERT OR REPLACE INTO {self.name}(_doc_id, cell_id, offset, time_at, _deleted) VALUES (?, ?, ?, ?, ?);'
        self._conn.execute(
            sql,
            (doc_id, cell_id, offset, time_now(), 0),
        )
        if commit:
            self._conn.commit()

    def bulk_add_address(
        self,
        doc_ids: List[str],
        cell_ids: Union[List[int], np.ndarray],
        offsets: Union[List[int], np.ndarray],
        commit: bool = True,
    ):
        sql = f'INSERT OR REPLACE INTO {self.name}(_doc_id, cell_id, offset, time_at, _deleted) VALUES (?, ?, ?, ?, ?);'
        self._conn.executemany(
            sql,
            [
                (doc_id, cell_id, offset, time_now(), 0)
                for doc_id, cell_id, offset in zip(doc_ids, cell_ids, offsets)
            ],
        )
        if commit:
            self._conn.commit()
